from pathlib import Path

import geopandas as gpd
import numpy as np
import pytest
from conftest import scenario_data
from numpy.testing import assert_almost_equal

from pywatemsedem.errors import (
    PywatemsedemVectorAttributeError,
    PywatemsedemVectorAttributeValueError,
)


class TestCreateModel:
    """Test class to test often used use cases in pywatemsedem model workflow.

    This class focuses on specific use cases that are assumed to be used in the API:

    - *test_all*: Create model input with base landuse, river vector, infrastructure
      vector, parcels vector and water vector.
    - *test_ommit_water* Create model input with base landuse, river vector,
      infrastructure vector, parcels vector with:
        - include grass strips vector
        - include technical tillage vector
    - *test_ommit_river* Create model input with base landuse, infrastructure vector,
       parcels vector and water vector
    - *test_ommit_inf* Create model input with base landuse, river vector, parcels
       vector and water vector
    - *test_ommit_parcels*: Create model input with base landuse, river vector,
       infrastructure vector and water vector.
    - *test_ommit_all*: Create model input with base landuse?

    The means to test equality of the rasters is the unique values and their number of
    occurence. Specically, the output C-factor, kTC and composite landuse rasters are
    tested as these are raster to which WaTEM/SEDEM is most senstive to. For the
    composite landuse raster, only number of agricultural pixels are tested, not the
    occurence of individual parcels.
    """

    @pytest.mark.saga
    def test_all(self, scenario):
        """Create WaTEM/SEDEM parcels raster with all possible input vectors/rasters"""
        scenario.vct_parcels = scenario_data.parcels
        scenario.prepare_cnws_model_input()

        # Composite land-use (test number of parcels pixels, and not unique id's)
        arr = scenario.composite_landuse.arr
        arr[arr > 0] = 1
        un, counts = np.unique(arr, return_counts=True)
        np.testing.assert_allclose(un, [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0])
        np.testing.assert_allclose(counts, [53, 4030, 1107, 3947, 534, 28236, 11537])

        # c-factor
        un, counts = np.unique(scenario.cfactor.arr, return_counts=True)

        np.testing.assert_allclose(un, np.array([0.0, 0.001, 0.01, 0.37]))
        np.testing.assert_allclose(counts, np.array([32737, 1111, 4056, 11540]))

        # kTC
        un, counts = np.unique(scenario.ktc.arr, return_counts=True)
        np.testing.assert_allclose(un, [-9.999e03, 1.000e00, 9.000e00, 9.999e03])
        np.testing.assert_allclose(counts, [28236, 5137, 11537, 4534])

        assert True

    @pytest.mark.saga
    def test_omit_water(self, scenario):
        """Omit water to create WaTEM/SEDEM parcels raster. This scenario is used standard
        in the initial development of pywatemsedem in Flanders."""
        scenario.vct_parcels = scenario_data.parcels
        scenario.catchm._vct_water = None

        # test prepare model input on C-factor, ktc and landuse-raster
        scenario.prepare_cnws_model_input()

        # Composite land-use (test number of parcels pixels, and not unique id's)
        arr = scenario.composite_landuse.arr
        arr[arr > 0] = 1
        un, counts = np.unique(arr, return_counts=True)
        np.testing.assert_allclose(un, [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0])
        np.testing.assert_allclose(counts, [2, 4056, 1111, 3965, 534, 28236, 11540])

        # c-factor
        un, counts = np.unique(scenario.cfactor.arr, return_counts=True)
        np.testing.assert_allclose(un, np.array([0.0, 0.001, 0.01, 0.37]))
        np.testing.assert_allclose(counts, np.array([32737, 1111, 4056, 11540]))

        # kTC
        un, counts = np.unique(scenario.ktc.arr, return_counts=True)
        np.testing.assert_allclose(un, [-9.999e03, 1.000e00, 9.000e00, 9.999e03])
        np.testing.assert_allclose(counts, [28236, 5167, 11540, 4501])

    @pytest.mark.saga
    def test_omit_parcels(self, scenario):
        """Omit parcels to create WaTEM/SEDEM parcels raster."""
        scenario.prepare_cnws_model_input()
        scenario.catchm._vct_water = None

        # Composite land-use (test number of parcels pixels, and not unique id's)
        arr = scenario.composite_landuse.arr
        arr[arr > 0] = 1
        un, counts = np.unique(arr, return_counts=True)
        np.testing.assert_allclose(un, [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0])
        np.testing.assert_allclose(counts, [2, 4056, 1111, 3965, 534, 28236, 11540])

        # c-factor
        un, counts = np.unique(scenario.cfactor.arr, return_counts=True)
        np.testing.assert_allclose(un, np.array([0.0, 0.001, 0.01, 0.37]))
        np.testing.assert_allclose(counts, np.array([32737, 1111, 4056, 11540]))

        # kTC
        un, counts = np.unique(scenario.ktc.arr, return_counts=True)
        np.testing.assert_allclose(un, [-9.999e03, 1.000e00, 9.000e00, 9.999e03])
        np.testing.assert_allclose(counts, [28236, 5167, 11540, 4501])

        assert True

    @pytest.mark.saga
    def test_add_grass_strips(self, scenario):
        """Test creating composite landuse-, C-factor-, kTC-raster for case without
        water, but with adding grass strips."""
        scenario.vct_parcels = scenario_data.parcels
        scenario.catchm._vct_water = None

        scenario.vct_grass_strips = scenario_data.grass_strips
        scenario.choices.dict_ecm_options["UseGras"] = 1

        # test prepare model input on C-factor, ktc and landuse-raster
        scenario.prepare_cnws_model_input()

        # Composite land-use (test number of parcels pixels, and not unique id's)
        arr = scenario.composite_landuse.arr
        arr[arr > 0] = 1
        un, counts = np.unique(arr, return_counts=True)
        np.testing.assert_allclose(un, [-6, -5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0])
        np.testing.assert_allclose(
            counts, [1529, 2, 3844, 1011, 3965, 534, 28236, 10323]
        )

        # c-factor
        un, counts = np.unique(scenario.cfactor.arr, return_counts=True)

        np.testing.assert_allclose(
            un,
            np.array(
                [
                    0.0,
                    0.001,
                    0.01,
                    0.028,
                    0.046,
                    0.082,
                    0.1,
                    0.118,
                    0.154,
                    0.19,
                    0.208,
                    0.262,
                    0.298,
                    0.37,
                ]
            ),
        )
        np.testing.assert_allclose(
            counts,
            np.array(
                [32737, 1011, 4164, 39, 199, 43, 201, 19, 433, 2, 15, 245, 13, 10323]
            ),
        )

        # kTC
        un, counts = np.unique(scenario.ktc.arr, return_counts=True)
        np.testing.assert_allclose(
            un,
            np.array(
                [
                    -9.99900000e03,
                    8.19012646e-01,
                    8.19048198e-01,
                    8.19448238e-01,
                    8.19700171e-01,
                    8.20093704e-01,
                    8.20708421e-01,
                    8.21668642e-01,
                    8.25511499e-01,
                    8.29171296e-01,
                    8.34888088e-01,
                    8.57767012e-01,
                    9.13591694e-01,
                    9.66757094e-01,
                    1.0,
                    1.38216332e00,
                    2.19312119e00,
                    9.00000000e00,
                    9.99900000e03,
                ]
            ),
        )
        np.testing.assert_allclose(
            counts,
            np.array(
                [
                    28236,
                    66,
                    27,
                    17,
                    189,
                    21,
                    39,
                    199,
                    43,
                    201,
                    19,
                    433,
                    2,
                    15,
                    4855,
                    245,
                    13,
                    10323,
                    4501,
                ]
            ),
        )

    @pytest.mark.saga
    def test_omit_river(self, scenario):
        """Omit river to create WaTEM/SEDEM parcels raster."""
        # TODO
        assert True

    @pytest.mark.saga
    def test_omit_infrastructure(self, scenario):
        """Omit infrastructure to create WaTEM/SEDEM parcels raster."""
        # TODO
        assert True

    @pytest.mark.saga
    def test_ommit_all(self, scenario):
        """Ommit all input sources to create WaTEM/SEDEM parcels raster, except base land-use
        raster"""
        # TODO
        assert True


class TestEndpoints:
    """Tests functionalities for endpoints"""

    @pytest.mark.saga
    def test_vct_endpoints(self, scenario):
        """Test vector endpoints assignment"""

        # test unique records and counts
        scenario.choices.dict_model_options["Include sewers"] = 1
        scenario.choices.dict_variables["SewerInletEff"] = 1
        scenario.vct_endpoints = scenario_data.endpoints

        # endpoints
        un, counts = np.unique(scenario.endpoints.arr, return_counts=True)
        np.testing.assert_allclose(un, [0.0, 1])
        np.testing.assert_allclose(counts, [49040, 404])

        # endpoints_ids
        un, counts = np.unique(scenario.endpoints_id.arr, return_counts=True)
        np.testing.assert_allclose(un, [0, 1.0, 2.0])
        np.testing.assert_allclose(counts, [48970, 394, 80])

    @pytest.mark.saga
    def test_vct_endpoints_efficiency(self, scenario, recwarn):
        """Test vector endpoints assignment with user-defined 'efficiency' column"""
        # feed a self-defined efficiency column (with 10 linestring to 75 %, remaining
        # NaN)
        df = gpd.read_file(scenario_data.endpoints)
        df["efficiency"] = 0.75
        scenario.choices.dict_variables["SewerInletEff"] = 1
        scenario.choices.dict_model_options["Include sewers"] = 1
        scenario.vct_endpoints = df
        un, counts = np.unique(scenario.endpoints.arr, return_counts=True)

        np.testing.assert_allclose(un, [0.0, 0.75])
        np.testing.assert_allclose(counts, [49040, 404])

        # feed a self-defined efficiency column (with all NaN)
        df = gpd.read_file(scenario_data.endpoints)
        df["efficiency"] = np.nan
        scenario.vct_endpoints = df
        un, counts = np.unique(scenario.endpoints.arr, return_counts=True)
        np.testing.assert_allclose(un, [0.0, 1.0])
        np.testing.assert_allclose(counts, [49040, 404])

        # catch warning from inputting all NaN-values to efficiency column
        w = recwarn.pop(UserWarning)
        assert (
            "The efficiency is not defined for all sewer line strings, assigning "
            "'SewerInletEff'-value" in str(w.message)
        )

    @pytest.mark.saga
    def test_vct_endpoints_efficiency_value_error(self, scenario):
        """Test wrong value in 'efficiency' attribute"""
        # feed negative efficiency values
        df = gpd.read_file(scenario_data.endpoints)
        df["efficiency"] = -0.1
        with pytest.raises(
            PywatemsedemVectorAttributeValueError,
            match="vector should contain values in",
        ):
            scenario.vct_endpoints = df

        df["efficiency"] = 10
        with pytest.raises(
            PywatemsedemVectorAttributeValueError,
            match="vector should contain values in",
        ):
            scenario.vct_endpoints = df

    @pytest.mark.saga
    def test_vct_endpoints_type_id_value_error(self, scenario):
        """Test wrong value in 'type_id' attribute"""
        # feed invalid type_id
        df = gpd.read_file(scenario_data.endpoints)
        df["type_id"] = np.nan
        with pytest.raises(
            ValueError, match="Please define a 'type_id' for every record."
        ):
            scenario.vct_endpoints = df


class TestParcels:
    @pytest.mark.saga
    def test_vct_parcels(self, scenario, recwarn):
        """Test assigment parcels"""
        # test assigment property
        scenario.vct_parcels = scenario_data.parcels

        # test type
        assert scenario.parcels.arr.dtype == np.int16
        assert scenario.parcels_ids.arr.dtype == np.float64

        # test number of values
        un = np.unique(scenario.parcels_ids.arr)
        np.testing.assert_allclose(len(un), 384)

        # test if parcels and parcels ids is equal IN CASE max id <= 32767
        np.testing.assert_allclose(scenario.parcels.arr, scenario.parcels_ids.arr)

        # manipulate parcels_ids and set above 32757 to test difference parcels_ids
        # and parcels getter
        df = gpd.read_file(scenario_data.parcels)
        df["NR"] = 1000**2

        scenario.vct_parcels = df
        assert np.max(scenario.parcels.arr) == 32767
        assert np.max(scenario.parcels_ids.arr) == 1000**2

        # catch warning
        w = recwarn.pop(UserWarning)
        assert (
            str(w.message) == "Parcels raster has values higher than the maximum "
            "allowed number for WaTEM/SEDEM definition (i.e. 32767). "
            "Setting values above 32767 to 32767."
        )

    @pytest.mark.saga
    def test_vct_parcels_value_error_landuse(self, scenario):
        """Test wrong value in 'LANDUSE' attribute parcels vector"""
        # feed incorrect values for column "LANDUSE"
        df = gpd.read_file(scenario_data.parcels)
        df["LANDUSE"] = -10
        with pytest.raises(
            PywatemsedemVectorAttributeValueError,
            match="vector can only contain values",
        ):
            scenario.vct_parcels = df

    @pytest.mark.saga
    def test_vct_parcels_attribute_error_landuse(self, scenario):
        """Test wrong attributes in parcels vector"""
        df = gpd.read_file(scenario_data.parcels)
        df = df.drop(columns="LANDUSE")
        with pytest.raises(
            PywatemsedemVectorAttributeError, match="input vector should contain"
        ):
            scenario.vct_parcels = df

    @pytest.mark.saga
    def test_vct_parcels_value_error_cfactor(self, scenario):
        """Test wrong value in 'C_crop' attribute parcels vector"""
        df = gpd.read_file(scenario_data.parcels)
        df["C_crop"] = -0.1
        with pytest.raises(
            PywatemsedemVectorAttributeValueError, match="vector should contain values"
        ):
            scenario.vct_parcels = df

    @pytest.mark.saga
    def test_vct_parcels_nan_value_cfactor(self, scenario):
        """Test if all nan in 'C_crop' is allowed"""
        df = gpd.read_file(scenario_data.parcels)
        df["C_crop"] = np.nan
        scenario.vct_parcels = df
        assert True


class TestGrassStrips:
    """Tests functionalities for grass strips:

    - source-oriented (grass strips, buffers)
    - symptom-oriented (tillage)
    """

    @pytest.mark.saga
    def test_vct_grass_strips(self, scenario, recwarn):
        """Test vector assignment"""
        scenario.choices.dict_ecm_options["UseGras"] = 0
        scenario.vct_grass_strips = scenario_data.grass_strips

        # set use gras to 1
        scenario.choices.dict_ecm_options["UseGras"] = 1

        # test output
        un, counts = np.unique(scenario.grass_strips.arr, return_counts=True)

        np.testing.assert_allclose(un, [-9999.0] + list(range(0, 115)))
        np.testing.assert_allclose(
            counts,
            np.array(
                [
                    47318,
                    7,
                    14,
                    27,
                    37,
                    26,
                    15,
                    3,
                    8,
                    13,
                    15,
                    27,
                    4,
                    10,
                    3,
                    28,
                    27,
                    15,
                    7,
                    11,
                    13,
                    22,
                    23,
                    20,
                    28,
                    14,
                    7,
                    6,
                    9,
                    5,
                    17,
                    5,
                    30,
                    20,
                    7,
                    4,
                    19,
                    108,
                    13,
                    25,
                    6,
                    10,
                    4,
                    15,
                    13,
                    10,
                    29,
                    26,
                    22,
                    8,
                    43,
                    12,
                    4,
                    11,
                    16,
                    1,
                    8,
                    8,
                    5,
                    7,
                    20,
                    12,
                    18,
                    19,
                    8,
                    29,
                    36,
                    5,
                    5,
                    16,
                    12,
                    29,
                    25,
                    32,
                    21,
                    25,
                    7,
                    8,
                    5,
                    30,
                    28,
                    23,
                    22,
                    17,
                    30,
                    3,
                    27,
                    11,
                    11,
                    23,
                    20,
                    10,
                    13,
                    25,
                    38,
                    2,
                    2,
                    15,
                    52,
                    26,
                    33,
                    10,
                    30,
                    32,
                    21,
                    18,
                    39,
                    9,
                    58,
                    53,
                    22,
                    25,
                    16,
                    20,
                    14,
                    16,
                ]
            ),
        )

        # Test if option 'UseGras' is set off
        scenario.choices.dict_ecm_options["UseGras"] = 0
        scenario.vct_grass_strips = scenario_data.grass_strips
        scenario.choices.dict_ecm_options["UseGras"] = 1

    @pytest.mark.saga
    def test_vct_grass_strips_width_value_error(self, scenario, tmp_path):
        """Test wrong value in 'width' attribute"""
        fname = Path(tmp_path) / "temp.shp"
        df = gpd.read_file(scenario_data.grass_strips)
        df["width"] = -10
        df.to_file(fname)
        with pytest.raises(
            PywatemsedemVectorAttributeValueError,
            match="vector should contain values in",
        ):
            scenario.vct_grass_strips = fname

    @pytest.mark.saga
    def test_vct_grass_strips_attribute_error(self, scenario, tmp_path):
        """Test wrong attributes"""
        fname = Path(tmp_path) / "temp.shp"
        df = gpd.read_file(scenario_data.grass_strips)
        df = df.drop(columns="width")
        df.to_file(fname)
        with pytest.raises(
            PywatemsedemVectorAttributeError, match="input vector should contain"
        ):
            scenario.vct_grass_strips = fname


class TestBuffers:
    """Tests functionalities for buffers
    #TODO: write test for buffer outlet
    """

    @pytest.mark.saga
    def test_vct_buffers(self, scenario, recwarn):
        """Test vector assignment"""
        scenario.choices.dict_ecm_options["Include buffers"] = 0
        scenario.vct_buffers = scenario_data.buffers
        # catch warning
        w = recwarn.pop(UserWarning)
        assert (
            str(w.message) == "Include buffers' in 'dict_ecm_options' is equal to 0. "
            "Will not include buffers."
        )

        # set include buffers on
        scenario.choices.dict_ecm_options["Include buffers"] = 1
        # test output
        un, counts = np.unique(scenario.buffers.arr, return_counts=True)
        np.testing.assert_allclose(un, [0, 1, 2, 16385, 16386])
        np.testing.assert_allclose(counts, [49437, 1, 1, 3, 2])

        # Test if option 'Include buffers' is set off
        scenario.choices.dict_ecm_options["Include buffers"] = 0
        scenario.vct_buffers = scenario_data.buffers
        w = recwarn.pop(UserWarning)  # make sure prev waring is popped
        assert scenario.buffers is None  # calling buffers raises warning
        w = recwarn.pop(UserWarning)
        assert (
            str(w.message) == "Option 'Include buffers' in erosion control measure"
            " options is 0, returning None"
        )
        scenario.choices.dict_ecm_options["Include buffers"] = 1

    def test_warning_buffers_overlap_rivers(self):
        """#TODO"""
        assert True

    @pytest.mark.saga
    def test_vct_buffers_eff_value_error(self, scenario):
        """Test wrong value in 'eff' attribute"""
        df = gpd.read_file(scenario_data.buffers)
        df["eff"] = -10
        with pytest.raises(
            PywatemsedemVectorAttributeValueError,
            match="vector should contain values in",
        ):
            scenario.vct_buffers = df

    @pytest.mark.saga
    def test_vct_buffers_attribute_error(self, scenario):
        """Test wrong attribute"""
        df = gpd.read_file(scenario_data.buffers)
        df = df.drop(columns="eff")
        with pytest.raises(
            PywatemsedemVectorAttributeError, match="input vector should contain"
        ):
            scenario.vct_buffers = df


class TestDitches:
    """Tests functionalities for ditches"""

    def test_vct_ditches(self):
        """
        #TODO
        """
        assert True


class TestConductiveDams:
    """Tests functionalities for conductive dams"""

    def test_vct_conductive_dams(self):
        """
        #TODO
        """
        assert True


class TestOutlets:
    """Tests functionalities for outlets"""

    @pytest.mark.saga
    def test_outlets(self, scenario):
        """Test vector assignment"""
        scenario.vct_outlets = scenario_data.outlets
        arr, counts = np.unique(scenario.outlets.arr, return_counts=True)
        np.testing.assert_allclose(arr, [-9999, 1, 2])
        np.testing.assert_allclose(counts, [49442, 1, 1])


class TestForceRouting:
    """Tests functionalities for force routing"""

    @pytest.mark.saga
    def test_forcerouting(self, scenario):
        """Test vector assignment"""
        scenario.vct_force_routing = scenario_data.force_routing
        df = scenario.vct_force_routing
        assert_almost_equal(
            df["fromX"].values, [163891.888238, 165939.062062], decimal=0
        )
        assert_almost_equal(df["toX"].values, [164574.279513, 166268.492332], decimal=0)
        assert_almost_equal(
            df["fromY"].values, [167144.040010, 167638.185416], decimal=0
        )
        assert_almost_equal(df["toY"].values, [167449.939547, 168320.576690], decimal=0)
        assert_almost_equal(df["fromcol"].values, [80, 182])
        assert_almost_equal(df["fromrow"].values, [119, 95])
        assert_almost_equal(df["torow"].values, [104, 60])
        assert_almost_equal(df["tocol"].values, [114, 199])



# seperate plotting funcion
import tempfile
import warnings
from dataclasses import dataclass
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd

# hvplot functionalities
from matplotlib import colors
from shapely.geometry import LineString

from pywatemsedem.plots import plot_cumulative_sedimentload

from ..geo.factory import Factory
from ..geo.utils import (
    check_raster_properties_raster_with_template,
    clean_up_tempfiles,
    create_spatial_index,
    execute_saga,
    load_raster,
    mask_array_with_val,
    raster_array_to_pandas_dataframe,
    raster_dataframe_to_arr,
    raster_to_polygon,
    rst_to_vct_points,
    write_arr_as_rst,
)
from ..io.plots import axes_creator  # hvplot_continuous_raster,
from ..io.plots import log_scale_enabler, plot_continuous_raster, plot_output_raster
from ..io.valid import valid_array_type, valid_boundaries, valid_non_nan

IMPLEMENTED_RASTER_TYPES = [np.int16, np.int32, np.int64, np.float32, np.float64]
COLORMAP = "cividis"
COLORMAP_SEDIOUT = colors.LinearSegmentedColormap.from_list(
    "SediOutcmap", ["#ffffff", "#fecc5c", "#ff8c00", "#d7191c", "#bd0026"]
)


@dataclass
class Modeloutput(Factory):
    def __init__(self, template, resolution, epsg, nodata):
        """AbstractRaster class with model outputs as attributes. Modeloutput class
        serves the goal of automating the reading in, checking and visualisation
        of the output data of the CNWS model.

        Parameters
        ----------
        template: pathlib.Path | str
            Template raster needed to extract raster properties.
            Raster contains values in [0,1]. Values in ]0,1] define the catchment.
        resolution: int
            See :class:`pywatemsedem.geo.RasterProperties`.
        epsg: int
            See :class:`pywatemsedem.geo.RasterProperties`.
        nodata: float
            See :class:`pywatemsedem.geo.RasterProperties`.
        """
        super().__init__(resolution, epsg, nodata, template.parent)
        self.mask = template

        # DATA
        self._sediout = None
        self._routing = None
        self._routing_missing = None
        self._ls = None
        self._slope = None
        self._uparea = None
        self._total_sediment = None
        self._sewer_in = None
        self._sedi_export = None
        self._sedi_in = None
        self._capacity = None
        self._rusle = None

    @property
    def sediout(self):
        """Getter SediOut attribute.

        For documentation, see :ref:`here <watemsedem:sedioutrst>`
        """
        return self._sediout

    @sediout.setter
    def sediout(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._sediout = self.raster_factory(raster, flag_mask=False)
        # checks on raster data
        valid_non_nan(self.sediout.arr)
        valid_array_type(self.sediout.arr, required_type=np.float32)
        valid_boundaries(
            self.sediout.arr[self.sediout.arr != -9999], lower=0, upper=None
        )

        title = "SediOut [kg/year]"

        def plot(
            fig=None, ax=None, ticks=[0, 10000, 20000, 40000, 80000], *args, **kwargs
        ):
            """Plot for SediOut with nonlinear colorscale. Masked under 100 kg/year.

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis
            ticks: list, default = [0,10000,20000,40000,80000]
                    Possibility to supply a list of 5 values for ticks of colorscale.
                    If ticks = None, 0th, 25th, 50th, 75th and 100th percentile of
                    the data are used as ticks

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            arr = np.ma.masked_where(self.sediout.arr < 100, self.sediout.arr)
            fig, ax = plot_output_raster(
                fig,
                ax,
                arr,
                self.mask.arr,
                self.rp.bounds,
                title,
                ticks,
                cmap=COLORMAP_SEDIOUT,
                *args,
                **kwargs,
            )
            return fig, ax

        self._sediout.plot = plot

    @property
    def routing(self):
        """Getter routing attribute.
        For documentation, see :ref:`here <watemsedem:routingtxt>`
        """
        return self._routing

    @routing.setter
    def routing(self, text):
        """Setter

        Parameters
        ----------
        text: pathlib.Path | str
        """

        self._routing = pd.read_table(text)
        self._txt_routing = text
        self.gdf_routing = None
        # checks
        valid_boundaries(self.routing["col"].values, lower=0, upper=self.rp.ncols)
        valid_boundaries(self.routing["row"].values, lower=0, upper=self.rp.nrows)

        arr_target1col = self.routing.loc[
            self.routing["target1col"] != -99, "target1col"
        ].values
        valid_boundaries(arr_target1col, lower=0, upper=self.rp.ncols)
        arr_target1row = self.routing.loc[
            self.routing["target1row"] != -99, "target1row"
        ].values
        valid_boundaries(arr_target1row, lower=0, upper=self.rp.nrows)
        valid_boundaries(self.routing["part1"].values, lower=0, upper=1)
        valid_boundaries(self.routing["distance1"], lower=0)

        arr_target2col = self.routing.loc[
            self.routing["target2col"] != -99, "target2col"
        ].values
        valid_boundaries(arr_target2col, lower=0, upper=self.rp.ncols)
        arr_target2row = self.routing.loc[
            self.routing["target2row"] != -99, "target2row"
        ].values
        valid_boundaries(arr_target2row, lower=0, upper=self.rp.nrows)
        valid_boundaries(self.routing["part2"].values, lower=0, upper=1)
        valid_boundaries(self.routing["distance2"].values, lower=0)

        def plot():
            """Interactive plot for routing

            Returns
            --------
            folium.folium.Map
            """
            if self.gdf_routing is None:
                msg = "Apply .make_routing_vector(...) on"
                msg1 = " Modeloutput before calling .plot()"
                raise IOError(msg + msg1)
            else:
                return self.gdf_routing.explore()
            # Future implementation: try to draw arrow in correct direction

        self._routing.plot = plot

    @property
    def routing_missing(self):
        """Getter routing_missing attribute.
        For documentation, see :ref:`here <watemsedem:missingroutingtxt>`
        """
        return self._routing_missing

    @routing_missing.setter
    def routing_missing(self, text):
        """Setter

        Parameters
        ---------
        text: pathlib.Path | str
        """
        self._routing_missing = pd.read_table(text)
        self._txt_routing_missing = text
        self.gdf_routing_missing = None

        if self.routing_missing.empty:
            warnings.warn("routing_missing is empty")
        else:
            # checks like routing
            valid_boundaries(
                self.routing_missing["col"].values, lower=0, upper=self.rp.ncols
            )
            valid_boundaries(
                self.routing_missing["row"].values, lower=0, upper=self.rp.nrows
            )

            arr_target1col = self.routing_missing.loc[
                self.routing["target1col"] != -99, "target1col"
            ].values
            valid_boundaries(arr_target1col, lower=0, upper=self.rp.ncols)
            arr_target1row = self.routing_missing.loc[
                self.routing["target1row"] != -99, "target1row"
            ].values
            valid_boundaries(arr_target1row, lower=0, upper=self.rp.nrows)
            valid_boundaries(self.routing_missing["part1"].values, lower=0, upper=1)
            valid_boundaries(self.routing_missing["distance1"], lower=0)

            arr_target2col = self.routing_missing.loc[
                self.routing["target2col"] != -99, "target2col"
            ].values
            valid_boundaries(arr_target2col, lower=0, upper=self.rp.ncols)
            arr_target2row = self.routing_missing.loc[
                self.routing["target2row"] != -99, "target2row"
            ].values
            valid_boundaries(arr_target2row, lower=0, upper=self.rp.nrows)
            valid_boundaries(self.routing_missing["part2"].values, lower=0, upper=1)
            valid_boundaries(self.routing_missing["distance2"].values, lower=0)

        def plot():
            """Interactive plot for routing_missing

            Returns
            --------
            folium.folium.Map
            """
            if self.routing_missing.empty:
                warnings.warn("routing_missing is empty, no plot is generated")
            else:
                if self.gdf_routing_missing is None:
                    msg = (
                        "Apply .make_routing_vector(...,routing_missing = True) on"
                        " Modeloutput before calling .plot()"
                    )
                    raise IOError(msg)
                else:
                    return self.gdf_routing_missing.explore()

        self._routing_missing.plot = plot

    @property
    def ls(self):
        """Getter LS (length-slope) attribute.

        For documentation, see :ref:`here <watemsedem:lsmap>`
        """
        return self._ls

    @ls.setter
    def ls(self, raster):
        """Setter

        Parameters
        ---------
        raster: pathlib.Path | str
        """
        self._ls = self.raster_factory(raster, flag_mask=False)

        valid_array_type(self.ls.arr, required_type=np.float32)
        valid_boundaries(self.ls.arr[self.ls.arr != -9999], lower=0, upper=None)
        check_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        title = "LS [-]"

        def plot(fig=None, ax=None, ticks=None, *args, **kwargs):
            """Plot for LS with non-linear colormap

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis
            ticks: list, default = None
                    Possibility to supply a list of 5 values for ticks of colorscale.
                    Default is 0th, 25th, 50th, 75th and 100th percentile

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            arr = mask_array_with_val(self.ls.arr, self.ls.arr, -9999)
            fig, ax = plot_output_raster(
                fig,
                ax,
                arr,
                self.mask.arr,
                self.rp.bounds,
                title,
                ticks,
                *args,
                **kwargs,
            )
            return fig, ax

        self._ls.plot = plot

    @property
    def slope(self):
        """Getter slope attribute.

        For documentation, see :ref:`here <watemsedem:slopemap>`
        """
        return self._slope

    @slope.setter
    def slope(self, raster):
        """Setter

        Parameters
        ---------
        raster: pathlib.Path | str
        """
        self._slope = self.raster_factory(raster, flag_mask=False)

        valid_array_type(self.slope.arr, required_type=np.float32)
        valid_boundaries(self.slope.arr[self.slope.arr != -9999], lower=0, upper=None)
        check_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        title = "Slope [rad]"

        def plot(fig=None, ax=None, ticks=None, *args, **kwargs):
            """Plot for Slope with non-linear colormap

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis
            ticks: list, default = None
                    Possibility to supply a list of 5 values for ticks of colorscale.
                    Default is 0th, 25th, 50th, 75th and 100th percentile of the data

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            arr = mask_array_with_val(self.slope.arr, self.slope.arr, -9999)
            fig, ax = plot_output_raster(
                fig,
                ax,
                arr,
                self.mask.arr,
                self.rp.bounds,
                title,
                ticks,
                *args,
                **kwargs,
            )
            return fig, ax

        self._slope.plot = plot

    @property
    def uparea(self):
        """Getter uparea attribute.

        For documentation, see :ref:`here <watemsedem:upareamap>`
        """
        return self._uparea

    @uparea.setter
    def uparea(self, raster):
        """Setter

        Parameters
        ---------
        raster: pathlib.Path | str
        """
        self._uparea = self.raster_factory(raster, flag_mask=False)

        valid_array_type(self.uparea.arr, required_type=np.float32)
        valid_boundaries(self.uparea.arr[self.uparea.arr != -9999], lower=0, upper=None)
        check_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        title = "uparea [m²]"

        def plot(fig=None, ax=None, ticks=None, *args, **kwargs):
            """Plot for uparea with non-linear colormap

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis
            ticks: list, default = None
                    Possibility to supply a list of 5 values for ticks of colorscale.
                    Default is 0th, 25th, 50th, 75th and 100th percentile

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            # masks out river based on ls
            if not self.ls:
                msg = "Assign ls to Modeloutput before plotting uparea"
                raise NotImplementedError(msg)
            else:
                arr = mask_array_with_val(self.uparea.arr, self.ls.arr, -9999)
            fig, ax = plot_output_raster(
                fig,
                ax,
                arr,
                self.mask.arr,
                self.rp.bounds,
                title,
                ticks,
                *args,
                **kwargs,
            )
            return fig, ax

        self._uparea.plot = plot

    @property
    def total_sediment(self):
        """Getter total sediment attribute.

        For documentation, see :ref:`here <watemsedem:totalsedimenttxt>`.
        For explanation on colmun variables of dataframe: see
        :func:`pywatemsedem.pywatemsedem.cnwsoutput.load_total_sediment_file`.
        """
        return self._total_sediment

    @total_sediment.setter
    def total_sediment(self, text):
        """Setter

        Parameters
        ----------
        text: pathlib.Path | str
        """
        dict = load_total_sediment_file(text)
        self._total_sediment = pd.DataFrame(dict, index=[0])

    @property
    def sewer_in(self):
        """Getter sewer_in attribute.

        For documentation, see :ref:`here <watemsedem:sewerinrst>`
        """
        return self._sewer_in

    @sewer_in.setter
    def sewer_in(self, raster):
        """Setter

        Parameters
        ---------
        raster: pathlib.Path | str
        """
        self._sewer_in = self.raster_factory(raster, flag_mask=True)

        valid_array_type(self.sewer_in.arr, required_type=np.float32)
        valid_boundaries(
            self.sewer_in.arr[self.sewer_in.arr != -9999], lower=0, upper=None
        )
        check_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)
        title = "sewer in [kg/year]"

        def plot(fig=None, ax=None, ticks=None, *args, **kwargs):
            """Plot for sewer_in with non-linear colormap. Zeros are filtered out

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis
            ticks: list, default = None
                    Possibility to supply a list of 5 values for ticks of colorscale.
                    Default is 0th, 25th, 50th, 75th and 100th percentile of
                    the data excluding zero values

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            arr = mask_array_with_val(self.sewer_in.arr, self.sewer_in.arr, -9999)
            arr = mask_array_with_val(arr, self.sewer_in.arr, 0)
            fig, ax = plot_output_raster(
                fig,
                ax,
                arr,
                self.mask.arr,
                self.rp.bounds,
                title,
                ticks,
                cmap=COLORMAP_SEDIOUT,
                *args,
                **kwargs,
            )
            return fig, ax

        self._sewer_in.plot = plot

    @property
    def sedi_export(self):
        """Getter SediExport attribute.

        For documentation, see :ref:`here <watemsedem:sediexportrst>`
        """
        return self._sedi_export

    @sedi_export.setter
    def sedi_export(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._sedi_export = self.raster_factory(raster, flag_mask=False)

        valid_array_type(self.sewer_in.arr, required_type=np.float32)
        valid_boundaries(
            self.sedi_export.arr[self.sedi_export.arr != -9999], lower=0, upper=None
        )
        check_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)
        title = "SediExport [kg/year]"

        def plot(fig=None, ax=None, ticks=None, *args, **kwargs):
            """Plot for SediExport with non-linear colormap. Zeros are filtered out

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis
            ticks: list, default = None
                    Possibility to supply a list of 5 values for ticks of colorscale.
                    Default is 0th, 25th, 50th, 75th and 100th percentile of the data
                    available in the rivercells excluding zeros

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(self.sedi_export.arr, self.mask.arr, 0)
            arr = mask_array_with_val(self.sedi_export.arr, self.sedi_export.arr, -9999)
            # mask where the river is not present
            if not self.ls:
                msg = "Assign ls to Modeloutput before plotting SediExport"
                raise NotImplementedError(msg)
            else:
                arr = np.ma.masked_where(self.ls.arr != -9999, arr)
            if not ticks:
                arr_nozeros = mask_array_with_val(arr, arr, 0)
                arr_nozeros = np.ma.filled(arr_nozeros, np.nan)
                ticks = np.nanpercentile(arr_nozeros, [0, 25, 50, 75, 100])
                ticks = np.round(
                    ticks
                ).tolist()  # avoids problems with invalid vmin or vmax
            fig, ax = plot_output_raster(
                fig,
                ax,
                arr,
                self.mask.arr,
                self.rp.bounds,
                title,
                ticks,
                cmap=COLORMAP_SEDIOUT,
                *args,
                **kwargs,
            )
            return fig, ax

        self._sedi_export.plot = plot

    @property
    def sedi_in(self):
        """Getter SediIn attribute.

        For documentation, see :ref:`here <watemsedem:sediinrst>`
        """
        return self._sedi_in

    @sedi_in.setter
    def sedi_in(self, raster):
        """Setter

        Parameters
        ----------
        raster: pathlib.Path | str
        """
        self._sedi_in = self.raster_factory(raster, flag_mask=False)

        valid_array_type(self.sedi_in.arr, required_type=np.float32)
        valid_boundaries(
            self.sedi_in.arr[self.sedi_in.arr != -9999], lower=0, upper=None
        )
        check_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)
        title = "SediIn [kg/year]"

        def plot(
            fig=None, ax=None, ticks=[0, 10000, 20000, 40000, 80000], *args, **kwargs
        ):
            """Plot for SediIn

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis
            ticks: list, default = [0,10000,20000,40000,80000]
                    Possibility to supply a list of 5 values for ticks of colorscale.
                    If ticks =None, 0th, 25th, 50th, 75th and 100th percentile
                    of the data are used as ticks

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            fig, ax = plot_output_raster(
                fig=fig,
                ax=ax,
                arr=self.sedi_in.arr,
                mask=self.mask.arr,
                title=title,
                bounds=self.rp.bounds,
                ticks=ticks,
                cmap=COLORMAP_SEDIOUT,
                *args,
                **kwargs,
            )

        self._sedi_in.plot = plot

    @property
    def capacity(self):
        """Getter capacity attribute.

        For documentation, see :ref:`here <watemsedem:capacityrst>`
        """
        return self._capacity

    @capacity.setter
    def capacity(self, raster):
        """Setter

        Parameters
        ---------
        raster: pathlib.Path | str
        """
        self._capacity = self.raster_factory(raster, flag_mask=False)

        valid_array_type(self.capacity.arr, required_type=np.float32)
        valid_boundaries(
            self.capacity.arr[self.capacity.arr != -9999], lower=0, upper=None
        )
        check_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        title = "Capacity [kg/year]"

        def plot(fig=None, ax=None, ticks=None, *args, **kwargs):
            """Plot for Capacity with logarithmic colormap

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis
            ticks: list, default = None
                    Possibility to supply a list of 5 values for ticks of colorscale.
                    Default is 0th, 25th, 50th, 75th and 100th percentile

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            if not self.ls:
                msg = "Assign ls to Modeloutput before plotting Capacity"
                raise NotImplementedError(msg)
            else:
                arr = mask_array_with_val(self.capacity.arr, self.ls.arr, -9999)
            fig, ax = axes_creator(fig, ax)
            arr = mask_array_with_val(arr, self.mask.arr, 0)
            arr = np.ma.filled(arr, np.nan)
            lower = log_scale_enabler(arr, cnorm="log")
            norm = colors.LogNorm(vmin=lower, vmax=np.nanmax(arr))
            fig, ax = plot_continuous_raster(
                fig,
                ax,
                arr=arr,
                bounds=self.rp.bounds,
                norm=norm,
                colorbar=True,
                ticks=ticks,
                *args,
                **kwargs,
            )
            ax.set_title(title)
            return fig, ax

        self._capacity.plot = plot

    @property
    def rusle(self):
        """Getter rusle attribute.


        For documentation, see :ref:`here <watemsedem:ruslerst>`
        """
        return self._rusle

    @rusle.setter
    def rusle(self, raster):
        """Setter

        Parameters
        ---------
        raster: pathlib.Path | str
        """
        self._rusle = self.raster_factory(raster, flag_mask=False)

        valid_array_type(self.capacity.arr, required_type=np.float32)
        valid_boundaries(
            self.capacity.arr[self.capacity.arr != -9999], lower=0, upper=None
        )
        check_raster_properties_raster_with_template(self.rp, raster, epsg=self.rp.epsg)

        title = "RUSLE [kg/(year.m²)]"

        def plot(fig=None, ax=None, ticks=None, *args, **kwargs):
            """Plot for RUSLE

            Parameters
            ----------
            fig: matplotlib.figure.Figure, default = None
                if not given, defaults to generating new figure
            ax: matplotlib.pyplot.axis, default = None
                if not given, defaults to generating new axis
            ticks: list, default = None
                    Possibility to supply a list of 5 values for ticks of colorscale.
                    Default is 0th, 25th, 50th, 75th and 100th percentile of the data
                    excluding zero values

            Returns
            -------
            fig: matplotlib.figure.Figure

            ax: matplotlib.axes.Axes
            """
            arr = mask_array_with_val(self.rusle.arr, self.rusle.arr, 0)
            fig, ax = plot_output_raster(
                fig=fig,
                ax=ax,
                arr=arr,
                mask=self.mask.arr,
                title=title,
                bounds=self.rp.bounds,
                ticks=ticks,
                *args,
                **kwargs,
            )

        self._rusle.plot = plot

    def make_routing_vector(
        self, landuse, sediout, percentile=90, routing_missing=False
    ):
        """Converts pandas dataframe of routing or routing_missing to a geopandas
        dataframe

        Parameters
        ----------
        landuse: pathlib.Path | str | str
                Path of landuse raster
        sediout: pathlib.Path | str | str
                Path of sediout raster
        percentile: int, default = 90
            Only vectors belonging to SediOutdata percentile and higher is kept
        routing_missing: bool, default = False
                        set to True to apply function to routing_missing instead
                        of routing

        Returns
        --------
        geopandas.GeoDataFrame
        """
        raster, landuse_rasterio_profile = load_raster(landuse)
        # cheks op raster
        valid_non_nan(raster)
        valid_array_type(raster, required_type=np.int16)
        valid_boundaries(raster, lower=-32757, upper=32757)
        check_raster_properties_raster_with_template(
            landuse, landuse, epsg=self.rp.epsg
        )

        # selecting what to vectorise
        sediout_arr, sediout_profile = load_raster(sediout)
        sediout_df = raster_array_to_pandas_dataframe(sediout_arr, sediout_profile)
        seditout_df_sel = sediout_df.loc[
            sediout_df["val"] > np.percentile(sediout_df["val"], percentile)
        ]
        if routing_missing:
            df = self.routing_missing
        else:
            df = self.routing
        df_temp = df.rename(
            columns={
                "target1col": "targetcol1",
                "target1row": "targetrow1",
                "target2col": "targetcol2",
                "target2row": "targetrow2",
            }
        )
        df_temp = pd.wide_to_long(
            df_temp,
            i=["col", "row"],
            stubnames=["targetcol", "targetrow", "distance", "part"],
            j="cell",
        ).reset_index()
        df_comb = seditout_df_sel.merge(df_temp, how="left", on=["row", "col"])
        # filter out data where part = 0 of None/NaN
        df_comb = df_comb[df_comb["part"] != 0]
        df_comb = df_comb[np.invert(pd.isnull(df_comb["part"]).values)]

        # + and - self.rp.resolution/2 for correction of vectors
        # being in center of cell
        df_comb["sourceX"] = (
            df_comb["col"] * self.rp.resolution
            + self.rp.bounds[0]
            - self.rp.resolution / 2
        )  # bounds[0] = xmin
        df_comb["sourceY"] = (
            -df_comb["row"] * self.rp.resolution
            + self.rp.bounds[3]
            + self.rp.resolution / 2
        )  # bounds[3] = ymax
        df_comb["targetX"] = (
            df_comb["targetcol"] * self.rp.resolution
            + self.rp.bounds[0]
            - self.rp.resolution / 2
        )
        df_comb["targetY"] = (
            -df_comb["targetrow"] * self.rp.resolution
            + self.rp.bounds[3]
            + self.rp.resolution / 2
        )
        geometry = [
            LineString(
                [
                    (df_comb.sourceX.iloc[i], df_comb.sourceY.iloc[i]),
                    (df_comb.targetX.iloc[i], df_comb.targetY.iloc[i]),
                ]
            )
            for i in range(len(df_comb))
        ]
        if routing_missing:
            self.gdf_routing_missing = gpd.GeoDataFrame(
                df_comb, geometry=geometry, crs=self.rp.epsg
            )
            return self.gdf_routing_missing
        else:
            self.gdf_routing = gpd.GeoDataFrame(
                df_comb, geometry=geometry, crs=self.rp.epsg
            )
            return self.gdf_routing


def get_prckrt_statistics(rst_prckrt, unit="ha", resolution=20):
    """Get the statistics of the WaTEM/SEDEM perceelskaart

    Load the WaTEM/SEDEM perceelskaart and compute for every unique id the area
    (ha or m2).

    Parameters
    ----------
    rst_prckrt: str or pathlib.Path | str
        File path ofWaTEM/SEDEM perceelskaart
    unit: str
        The unit of the output, currently pixel, m2 and ha implemented
    resolution: int
        Resolution of the rasters

    Returns
    -------
    out: dict,
        Area for each class, keys:

            - *agriculture* (float): area (ha/m2)
            - *river* (float): area (ha/m2)
            - *infrastructure* (float): area (ha/m2)
            - *forest* (float): area (ha/m2)
            - *meadow* (float): area (ha/m2)
            - *open_water* (float): area (ha/m2)
            - *gras_strip* (float): area (ha/m2)
    """
    arr_prckrt, profile = load_raster(rst_prckrt, list_format=False)
    arr_prckrt = arr_prckrt[arr_prckrt != profile["nodata"]]
    arr_prckrt[arr_prckrt > 0] = 1
    unique, counts = np.unique(arr_prckrt, return_counts=True)

    if unit in ["ha", "pixel", "m2"]:
        if unit == "ha":
            convert_unit = resolution**2 / 100**2
        elif unit == "pixel":
            convert_unit = 1
        elif unit == "m2":
            convert_unit = resolution**2
    else:
        msg = (
            f"Unit '{unit}' not defined, cannot convert unit from pixel to "
            "unit. Please select pixel, ha or m2"
        )
        return IOError(msg)

    out = {}

    out["agriculture"] = counts[unique == 1][0] if 1 in unique else 0
    out["river"] = counts[unique == -1][0] if -1 in unique else 0
    out["infrastructure"] = counts[unique == -2][0] if -2 in unique else 0
    out["forest"] = counts[unique == -3][0] if -3 in unique else 0
    out["meadow"] = counts[unique == -4][0] if -4 in unique else 0
    out["open_water"] = counts[unique == -5][0] if -5 in unique else 0
    out["gras_strip"] = counts[unique == -6][0] if -6 in unique else 0

    for i in list(out.keys()):
        if convert_unit != 1:
            out[i] = out[i] * convert_unit

    return out


def make_routing_vct_saga(
    txt_routing, rst_prckrt, vct_out, rstparams, extent=None, tile_number=None
):
    """Generate a routing vct routing file (added with CNWS landuse) based on
    the routing table
    extent and tilenumber are defined to make a routing file only for a
    certain extent

    Parameters
    ----------
    txt_routing: str or pathlib.Path | str
        File path of the WaTEM/SEDEM routing tabl
    rst_prckrt: str
        name of CNWS input 'perceelskaart'
    vct_out:
        name of the shape outputfile
    rstparams: dict
        gdal dictionary holding all metadata for idrisi rasters
    extent: list, default None
        list holding value of extent to consider, xmin,ymin,xmax,ymax
    tile_number: int, optional, default None
        number of the tile, used to add to filename

    Returns
    -------
    vct_out: str
        The resulting vector file will contain following features:

        - *col* (int): source column number.
        - *row* (int): source row number.
        - *targetcol* (int): target column number.
        - *targetcol* (int): source column number.
        - *part* (float): fraction of source that flows to target (between 0 and 1).
        - *distance* (float): distance of routing (m)
        - *landuSource* (int): landuse code of source pixel
        - *landuTarg* (int): landuse code of target pixel
        - *jump* (int): yes (1) or no (0)
        - *targetX* (float): x coor value of target
        - *targetY* (float): y coor value of target
        - *sourceX* (float): x coor value of source
        - *sourceY* (float): y coor value of source
    """

    # condition tag for making routing shape
    if txt_routing.exists():
        # check if file is tab seperated
        with open(txt_routing) as f:
            first_line = f.readline()
        seperator = ";" if "\t" not in first_line else "\t"
        # prepare the txt_file for conversion to routing shape
        vct_temp, condition = prepare_make_routing_vct_saga(
            txt_routing, vct_out, seperator, rstparams, extent, tile_number
        )

        # if condition True, run, if False (no routing segments available in
        # routing table) don't run
        if condition:
            run_saga_make_routing_shp_cmd(txt_routing, rst_prckrt, vct_out)
            return vct_out
        else:
            msg = "No valid extent to clip routing "
            raise Warning(msg)
    else:
        msg = f"'{txt_routing}' does not exist!"
        raise IOError(msg)


def prepare_make_routing_vct_saga(
    txt_routing, vct_out, seperator, rstparams, extent, tile_number
):
    """
    Prepare the input files for make routing vct_saga
    If an extension is defined, then this function will clip the file to
    this rectangular geographical extent

    Parameters
    ----------
    txt_routing_nonriver: str or pathlib.Path | str
        File path of the WaTEM/SEDEM routing tabl
    vct_out: str
        name of output shapefile for 'run_saga_make_routing_shp_cmd' fuction
    seperator: str
        current delimited of txt_routing
    rstparams: dict
        gdal dictionary holding all metadata for idrisi rasters
    extent: list
        min and max of rectangular frame to clip df [xmin, xmax, ymin, ymax]
    tile_number: int
        number of tile

    Returns
    -------
    vct_temp: str
        The resulting vector file will contain following features:

        - *col* (int): source column number.
        - *row* (int): source row number.
        - *targetcol* (int): target column number.
        - *targetcol* (int): source column number.
        - *part* (float): fraction of source that flows to target (between 0 and 1).
        - *distance* (float): distance of routing (m)
        - *landuSource* (int): landuse code of source pixel
        - *landuTarg* (int): landuse code of target pixel
        - *jump* (int): yes (1) or no (0)
        - *targetX* (float): x coor value of target
        - *targetY* (float): y coor value of target
        - *sourceX* (float): x coor value of source
        - *sourceY* (float): y coor value of source
    condition: boolean
        True if make_routing_shp has to be run

    """
    # check if an extent is defined for conversion to shape file, if not,
    # take whole domain txt_routing
    condition = True

    if extent is not None:
        df_routing = open_txt_routing_file(txt_routing)
        if df_routing.empty:
            msg = f"{txt_routing.name} is empty"
            raise IOError(msg)
        df_routing = condition_routing_dataframe_on_extent(
            df_routing, rstparams, extent
        )
        # rename files
        tag = f"_selected_{(0 if tile_number is None else tile_number)}"
        txt_routing = txt_routing.parent / txt_routing.stem + tag + ".txt"
        vct_temp = vct_out.parent / vct_out.stem + tag + ".shp"
        if seperator != "\t":
            df_routing.to_csv(txt_routing, sep="\t", index=False)
        # condition true if routing directions are found within extent
        condition = False if len(df_routing) == 0 else True

    else:
        # condition = True if no extent given
        vct_temp = vct_out
        condition = True
        df_routing = open_txt_routing_file(txt_routing)
        if df_routing.empty:
            msg = f"{txt_routing.name} is empty"
            raise IOError(msg)
        if seperator != "\t":
            df_routing.to_csv(txt_routing, sep="\t", index=False)

    return vct_temp, condition


def condition_routing_dataframe_on_extent(df_routing, rstparams, extent):
    """
    Clip routing_dataframe on defined extent

    Parameters
    ----------
    df_routing: pandas.DataFrame
        See :func:`pywatemsedem.io.modeloutput.open_txt_routing_file`
    rstparams: dict
        gdal dictionary holding all metadata for idrisi rasters
    extent: list
        min and max of rectangular frame to clip df [xmin, xmax, ymin, ymax]

    Returns
    -------
    df_routing: pandas.DataFrame
        See :func:`pywatemsedem.io.modeloutput.open_txt_routing_file`

    """
    # get xmin and ymin from rasterproperties
    xmin, ymin = (
        rstparams["minmax"][0],
        rstparams["minmax"][1],
    )
    # compile x and y's of source points in df
    df_routing["x"] = xmin + df_routing["col"] * rstparams["res"]
    df_routing["y"] = ymin + (rstparams["nrows"] - df_routing["row"]) * rstparams["res"]
    # condition
    cond = (
        (df_routing["x"] > extent[0])
        & (df_routing["x"] < extent[1])
        & (df_routing["y"] > extent[2])
        & (df_routing["y"] < extent[3])
    )
    df_routing = df_routing.loc[cond]

    return df_routing


def run_saga_make_routing_shp_cmd(txt_routing, rst_prckrt, vct_out):
    """
    Run the saga make routing shape command. This command makes from a pywatemsedem
    routing table and a pywatemsedem 'perceelskaart' a routing shapfile
    representation.

    The resulting vector file will contain following features:

    - *col* (int): source column number.
    - *row* (int): source row number.
    - *targetcol* (int): target column number.
    - *targetcol* (int): source column number.
    - *part* (float): fraction of source that flows to target (between 0 and 1).
    - *distance* (float): distance of routing (m)
    - *landuSource* (int): landuse code of source pixel
    - *landuTarg* (int): landuse code of target pixel
    - *jump* (int): yes (1) or no (0)
    - *targetX* (float): x coor value of target
    - *targetY* (float): y coor value of target
    - *sourceX* (float): x coor value of source
    - *sourceY* (float): y coor value of source

    Parameters
    ----------
    txt_routing_nonriver: str or pathlib.Path | str
        File path of the WaTEM/SEDEM routing table
    rst_prckrt: str
        File path of the WaTEM/SEDEM 'perceelskaart' raster
    vct_out: str
        File path of the output file


    """
    cmd_args = ["saga_cmd", "-f=s", "topology", "2"]
    cmd_args += ["-ROUTING", str(txt_routing)]
    cmd_args += [
        "-LANDUSE",
        str(rst_prckrt),
    ]
    cmd_args += ["-OUTPUTLINES", str(vct_out)]
    execute_saga(cmd_args)
    create_spatial_index(vct_out)


def define_subcatchments_saga(
    rst_in,
    txt_routing,
    resmap,
    rasterprop,
    tag="",
):
    """
    Define subcatchment for several points defined with a unique id in the
    raster

    Parameters
    ----------
    rst_in:
        Raster with single-pixels for which subcatchment should be determined
        > 0: determine subcatchment
        = 0: don't determine subcatchments
    txt_routing_nonriver: str or pathlib.Path | str
        File path of the WaTEM/SEDEMrouting table
    resmap: str or pathlib.Path | str
        Folder path where results should be saved
    rasterprop: dict
        rasterio profile dictionary holding all metadata for geotiff rasters
    tag: str or pathlib.Path | str
        File path of output shape and raster
    catchment_name: str
        Name of catchment, default ""
    scenario_label: str, default ""
        Scenario number or letter.

    Returns
    -------
    rst_subcatchments: str or pathlib.Path | str
        File path of raster with pixels beloging to a subcatchment having id
        equal to id sink in rst_in
    vct_subcatchments: str or pathlib.Path | str:
        File path of shapefile with polygon being the subcatchment having an
        id equal         to id sink in rst_in
    """
    # txt = os.path.join(self.scenario.outfolder, 'routing.txt')
    rst_subcatchments = resmap / f"subcatchments_{tag}.sdat"
    vct_subcatchments = rst_subcatchments.with_suffix(".shp")

    startvals = np.unique(load_raster(rst_in)[0])
    if len(startvals) <= 1:
        msg = "No values in input grid to define subcatchments!"
        raise ValueError(msg)

    cmd_args = ["saga_cmd", "-f=s", "topology", "3", "-ROUTING", str(txt_routing)]
    cmd_args += ["-SEGMENTS", str(rst_in)]
    cmd_args += ["-CATCH", str(rst_subcatchments)]

    execute_saga(cmd_args)
    raster_to_polygon(rst_subcatchments, vct_subcatchments)
    create_spatial_index(vct_subcatchments)
    gdf_subcatchments = gpd.read_file(vct_subcatchments)
    gdf_subcatchments.drop(columns=["ID", "NAME"], inplace=True)
    gdf_subcatchments["VALUE"] = gdf_subcatchments["VALUE"].astype("int32")
    gdf_subcatchments["AREA_HA"] = gdf_subcatchments.area / 10000.0
    gdf_subcatchments = gdf_subcatchments.set_crs(rasterprop["epsg"])
    gdf_subcatchments.to_file(vct_subcatchments)
    create_spatial_index(vct_subcatchments)

    return (
        (rst_subcatchments.parent / (rst_subcatchments.stem + ".sdat")),
        vct_subcatchments,
    )


def identify_individual_priority_catchments(
    arr_sediout,
    rst_profile,
    txt_routing_non_river,
    nmax,
    resmap=Path.cwd(),
    epsg="",
):
    """
    identify the individual priority catchments and add them to a raster
    and shapes dictionary

    Parameters
    ----------
    arr_sediout: numpy.ndarray
        numpy array format of sedout raster
    rst_profile: rasterio profile
        rasterio profile of the sedout raster
    txt_routing_nonriver: str or pathlib.Path | str
        File path of the WaTEM/SEDEM routing table
    nmax: int
        maximum number of catchment

    Returns
    -------
    subcatchmentpriority: geopandas.GeoDataFrame
        catchment shapes with number of catchment
    """

    n = 1
    id_ = 0
    # loop until break is encountered (no more pixels to cluster,
    # or a maximum number of clusters is reached)
    while True:

        # identify point with highest sediment load
        id_, max_sediout = create_id_raster_for_highest_value_arr(
            arr_sediout, id_, rst_profile, resmap=resmap / "priority_catchments"
        )

        # identify subcatchment/cluster coupled to this point
        template_name = (
            resmap / "priority_catchments" / f"subcatchments_priority_{n}.shp"
        )
        if not template_name.exists():
            rst_subcatch, vct_subcatch = define_subcatchments_saga(
                id_, txt_routing_non_river, tag=template_name
            )
            # assign sediout value to self.subcatchmprioritSHP
            gdf = gpd.read_file(vct_subcatch)
            gdf["sediout"] = max_sediout
            gdf.to_file(vct_subcatch)
        else:
            vct_subcatch = template_name
            rst_subcatch = vct_subcatch.with_suffix(".sdat")

        # condition: if a max number of clusters is identified
        # OR all pixels are classified: stop
        if (n >= nmax) | (np.sum(arr_sediout != rst_profile["nodata"]) == 0):
            nmax = n
            break
        else:
            arr_subcatch, _ = load_raster(rst_subcatch)
            arr_sediout[arr_subcatch != -99999.0] = rst_profile["nodata"]
            n += 1

    # merge different subcatchments to one file
    lst_gdf = []
    for i in Path("priority_catchments").iterdir():
        if i.suffix == ".shp":
            lst_gdf.append(gpd.read_file(i))
    gdf_subcatchmpriority = pd.concat(lst_gdf)

    gdf_subcatchmpriority.crs = {"init": epsg}
    dst = resmap / "priority_catchments.shp"
    gdf_subcatchmpriority.to_file(dst)
    create_spatial_index(dst)

    return gdf_subcatchmpriority


def create_id_raster_for_highest_value_arr(arr, id_, profile, resmap):
    """Create a raster with an id value assigned to the highest value in the raster"

    Parameters
    ----------
    arr: str or pathlib.Path | str
        with floats
    id_: int
        Sequential number of the catchment
    profile: rasterio profile
        Rasterio profile of the sedout raster
    resmap: str or pathlib.Path | str, optional
        Folder path to write results to

    Returns
    -------
    rst_id: str
        File path of the raster with id for highest value in raster
    val: float
        Maximum value in raster
    """
    resmap = Path(resmap)
    if not resmap.exists():
        (resmap).mkdir(parents=True, exist_ok=True)

    arr_id = arr.copy()
    max_val = np.max(arr)
    cond = arr == max_val
    arr_id[cond] = id_
    arr_id[~cond] = profile["nodata"]

    # write to disk
    rst_id = resmap / f"id_{id_}.rst"
    write_arr_as_rst(arr_id, rst_id, "int32", profile)

    return rst_id, max_val


def remove_river_routing(df_routing):
    """Remove river routing from routing dataframe

    Parameters
    ----------
    df_routing: pandas.DataFrame
        Long format of routing table

    Returns
    -------
    df_routing
        Long format of routing table, with river routing removed
    """
    df_routing = df_routing.loc[(df_routing["lnduSource"] != -1)]
    cols = [
        "lnduSource",
        "lnduTarg",
        "jump",
        "geometry",
        "targetX",
        "targetY",
        "sourceX",
        "sourceY",
    ]
    df_routing.drop(columns=cols, inplace=True)

    return df_routing


def compute_efficiency_buffers(rst_buffer, rst_sediin, rst_sediout):
    """Compute efficiency per buffer

    This function calculates the incoming and outgoing sediment per buffer.
    The deposition is computed by substracting the outgoing from the ingoing
    sediment.

    Parameters
    ----------
    rst_buffer: str or pathlib.Path | str
        File path of buffer raster with buffer id's
    rst_sediin: str or ppathlib.Path | str
        File path of CNWS SediIn raster, incoming sediment per pixel
    rst_sediout: str or pathlib.Path | str
        File path of CNWS SediOut raster, outgoing sediment per pixel

    Returns
    -------
    df_output: pandas.DataFrame
        Holding results of mass balance of buffers.

        - *buf_id* (float): id of the buffer (as in the buffer raster)
        - *sediin* (float): total incoming sediment in the buffer.
        - *sediout* (float): total outgoing sediment in the buffer.
        - *buff_sed* (float): amount sediment deposited in the buffer.
    """
    arr_buffers, profile = load_raster(rst_buffer)
    arr_sediin, _ = load_raster(rst_sediin)
    arr_sediout, _ = load_raster(rst_sediout)

    df_out = pd.DataFrame(columns=["NR", "sediin", "sediout", "buff_sed"])
    condition = (arr_buffers < 2**14) & (arr_buffers > 0)
    arr_unique_buffer = np.unique(arr_buffers[condition])
    for buf_id in arr_unique_buffer:
        arr_sediin_buffer = np.sum(arr_sediin[arr_buffers == buf_id])
        arr_sediout_buffer = np.sum(arr_sediout[arr_buffers == buf_id])
        arr_deposition = arr_sediin_buffer - arr_sediout_buffer
        df_out.loc[df_out.shape[0] + 1] = [
            buf_id,
            arr_sediin_buffer,
            arr_sediout_buffer,
            arr_deposition,
        ]
    return df_out


def open_txt_routing_file(txt_routing):
    """Open routing file with exceptions and seperators as needed

    Parameters
    ----------
    txt_routing: str or pathlib.Path | str
        File path of the WaTEM/SEDEM routing table

    Returns
    -------
    df_routing: pandas.DataFrame
        The routing DataFrame contains the following columns:

        - *col* (int): source column in rasters
        - *row* (int): source row in rasters
        - *target1col* (int): first target column in rasters
        - *target1row* (int): first target row in rasters
        - *part1* (float): Part that flows to target1
        - *distance1* (float): Distance of 1st vector
        - *target2col* (int): second target column in rasters
        - *target2row* (int): second target row in rasters
        - *part2* (float): Part that flows to target2
        - *distance2* (float): Distance of 2nd vector
    """
    try:
        with open(txt_routing) as f:
            first_line = f.readline()
        seperator = ";" if "\t" not in first_line else "\t"
        df_routing = pd.read_csv(txt_routing, sep=seperator)
        if df_routing.empty:
            msg = f"{txt_routing.name} is empty"
            raise IOError(msg)
    except IOError:
        msg = f"'{txt_routing}' does not exist"
        return IOError(msg)

    return df_routing


def create_erosion_raster(rst_watereros):
    """Create erosion raster from watereros.

    Extract negative values from Watereros raster and set positives to 0. Erosion
    values in this raster are positive (inverse of watereros)

    Parameters
    ----------
    rst_watereros: str or pathlib.Path | str
        File path to WaTEM/SEDEM watereros raster.

    Returns
    -------
    rst_erosion: pathlib.Path | str
        File path to pywatemsedem erosion raster.
    """
    rst_erosion = Path(rst_watereros).parent / "Erosion (kg).rst"
    if not rst_erosion.exists():
        arr_watereros, profile = load_raster(rst_watereros)
        arr_watereros[arr_watereros > 0] = 0
        arr_watereros[arr_watereros != profile["nodata"]] = -arr_watereros[
            arr_watereros != profile["nodata"]
        ]
        write_arr_as_rst(arr_watereros, rst_erosion, arr_watereros.dtype, profile)

    return rst_erosion


def create_deposition_raster(rst_watereros):
    """Create deposition raster from watereros.

    Extract positve values from Watereros raster and set negatives to 0

    Parameters
    ----------
    rst_watereros: str or pathlib.Path | str
        File path to WaTEM/SEDEM watereros raster.

    Returns
    -------
    rst_deposition: pathlib.Path | str
        File path to pywatemsedem erosion raster.
    """
    rst_deposition = Path(rst_watereros).parent / "Deposition (kg).rst"
    if not rst_deposition.exists():
        arr_watereros, profile = load_raster(rst_watereros)
        arr_watereros[arr_watereros < 0] = 0
        arr_watereros[arr_watereros != profile["nodata"]] = arr_watereros[
            arr_watereros != profile["nodata"]
        ]
        write_arr_as_rst(arr_watereros, rst_deposition, arr_watereros.dtype, profile)

    return rst_deposition


def map_rank_sediment_loads(
    rst_sediexport, threshold, vct_out="rank.shp", rst_endpoints=None, unit="kg"
):
    """Rank the sediment loads in sediexport (and sewerin) from high to low

    This function uses the rank output raster/dataframe of
    :func:`pywatemsedem.io.modeloutput.identify_rank_sediment_loads`
    to map a vector point file with information on rank, sediment output.

    For computation details, see
    :func:`pywatemsedem.io.modeloutput.identify_rank_sediment_loads`

    Parameters
    ----------
    rst_sediexport: str or pathlib.Path | str
        File path of WaTEM/SEDEM sediexport raster.
    threshold: float
        See :func:`pywatemsedem.io.modeloutput.compute_cumulative_loads_in_sinks`
    vct_out: str or pathlib.Path | str
        File path of output vector.
    rst_endpoints: str or pathlib.Path | str, default None
        File path to WaTEM/SEDEM sewerin raster (endpoints in pywatemsedem).
    unit: str
        "kg" or "ton"

    Note
    ----
    Only points are taken into account that account for x of the highest sediment loads.
    x is determined by threshold.
    """
    if unit not in ["kg", "ton"]:
        f"Unit '{unit}' should be either 'kg' op 'ton'."

    rst_out = tempfile.NamedTemporaryFile(suffix=".rst").name
    df_sediexport, threshold = identify_rank_sediment_loads(
        rst_sediexport, threshold, rst_out, rst_endpoints
    )
    rst_to_vct_points(rst_out, vct_out)
    gdf_out = gpd.read_file(vct_out)
    gdf_out["rank"] = gdf_out[Path(rst_out).stem]
    gdf_out = gdf_out.merge(df_sediexport, on="rank")
    if unit == "ton":
        gdf_out["sediexport"] = gdf_out["sediexport"] / 1000
    clean_up_tempfiles(Path(rst_out), "rst")

    return gdf_out


def identify_rank_sediment_loads(
    rst_sediexport, threshold, rst_out, rst_endpoints=None
):
    """Identify the highest ``threshold`` percentage sediment loads.

    This functions identifies the cumulative distribution of the sediexport
    (and sewerin, optional) raster:

    - (optional) map sewerin  raster to sediexport raster
    - convert sediexport raster to a list-format
    - sort from high to low
    - compute cumulative distribution
    - classify

    Parameters
    ----------
    rst_sediexport: str or pathlib.Path | str
        File path to WaTEM/SEDEM sediexport raster.
    threshold: float
        See :func:`pywatemsedem.io.modeloutput.compute_cumulative_loads_in_sinks`
    rst_out: str of pathlib.Path | str
        Output raster containing ranks of highest sediment loads (1: highest,
        2: second highest, ..)
    rst_endpoints: str or pathlib.Path | str
        File path to WaTEM/SEDEM sewerin raster.

    Returns
    -------
    df_sediexport: pandas.DataFrame
        Data Frame format of SediExport raster (format: see
        :func:`pywatemsedem.pywatemsedem.utils.raster_array_to_pandas_dataframe`)
    threshold: float
        See :func:`pywatemsedem.io.modeloutput.compute_cumulative_loads_in_sinks`
    """
    arr_sediexport, profile = load_raster(rst_sediexport)
    arr_sediexport = np.where(arr_sediexport == profile["nodata"], 0, arr_sediexport)

    if rst_endpoints is not None:
        arr_sewerin, _ = load_raster(rst_endpoints)
        arr_sewerin = np.where(arr_sewerin == profile["nodata"], 0, arr_sewerin)
        arr_sediexport += arr_sewerin

    df_sediexport = raster_array_to_pandas_dataframe(arr_sediexport, profile)
    profile["driver"] = "GTiff"

    # sort and select points
    df_sediexport, threshold = compute_cumulative_loads_in_sinks(
        df_sediexport, profile, threshold, plot=False
    )
    arr_sediexport = raster_dataframe_to_arr(df_sediexport, profile, "rank", np.float32)

    write_arr_as_rst(
        arr_sediexport,
        rst_out,
        "float32",
        profile,
    )

    return df_sediexport, threshold


def compute_cumulative_loads_in_sinks(
    df_sediexport, profile, threshold, delta=10, plot=False
):
    """Analyse cumulative sediment load by sorting SediExport values
    from high to low

    Parameters
    ----------
    df_sediexport: pandas.DataFrame
        Data Frame format of SediExport raster (format: see
        :func:`pywatemsedem.pywatemsedem.utils.raster_array_to_pandas_dataframe`)
    profile: rasterio.profiles
        See :func:`rasterio.open`
    threshold: int
        x percentage highest load that the user wants to analyse
    delta: int
        Delta used to iterate percentage
    plot: bool, default False
        True if you want a cumulative SediExport plot

    Returns
    -------
    df_sediexport: pandas.DataFrame
        Data Frame format of SediExport raster (format: see
        :func:`pywatemsedem.pywatemsedem.utils.raster_array_to_pandas_dataframe`) added
        with:

        - *cum_perc* (float): cumulative highest load
        - *perc* (float): percentage highest load
        - *class* (int): class as defined by `delta_perc`

    percentage: str
        updated percentage
    """

    # sort according to values of sediment load into river
    df_sediexport["sediexport"] = df_sediexport["val"]
    df_sediexport = df_sediexport.sort_values("sediexport", ascending=False)

    # calculate cumulative sum, in percentage
    cond = (df_sediexport["sediexport"] != profile["nodata"]) & (
        df_sediexport["val"] != 0.0
    )
    df_sediexport.loc[cond, "cum_sum"] = df_sediexport.loc[cond, "sediexport"].cumsum()
    df_sediexport.loc[cond, "cum_perc"] = (
        100
        * df_sediexport.loc[cond, "cum_sum"]
        / df_sediexport.loc[cond, "sediexport"].sum()
    )

    if plot:
        plot_cumulative_sedimentload(
            df_sediexport.loc[cond],
            threshold,
            "cumulative_sediexport.png",
        )

    # hotfix on percentage: if the first percentage is higher than the
    # user-predefined percentage, adjust it (small catchments)!
    threshold = verify_highest_load_with_threshold(df_sediexport, threshold)

    # prepare ids for subcatchment delineation
    df_sediexport["rank"] = profile["nodata"]
    df_sediexport["class"] = profile["nodata"]

    # assign unique id's - in order of importance - to records
    cond = (df_sediexport["cum_perc"] <= threshold) & (
        ~df_sediexport["cum_perc"].isnull()
    )
    df_sediexport.loc[cond, "rank"] = np.arange(np.sum(cond)) + 1

    # calculate percentage
    df_sediexport["perc"] = [
        df_sediexport["cum_perc"].iloc[i] - df_sediexport["cum_perc"].iloc[i - 1]
        if i != 0
        else df_sediexport["cum_perc"].iloc[i]
        for i in range(0, len(df_sediexport))
    ]

    # chekc if begin percentage is below delta_perc
    bperc = delta
    eperc = int(threshold + 1)
    if df_sediexport["cum_perc"].iloc[0] > bperc:
        bperc = int(np.ceil(df_sediexport["cum_perc"].iloc[0] / 10) * 10)

    for i in range(bperc, eperc, delta):
        cond = (
            (df_sediexport["cum_perc"] > i - delta)
            & (df_sediexport["cum_perc"] <= i)
            & (~df_sediexport["cum_perc"].isnull())
        )
        df_sediexport.loc[cond, "class"] = i

    return (
        df_sediexport[
            ["col", "row", "rank", "perc", "cum_perc", "class", "sediexport"]
        ],
        int(threshold),
    )


def verify_highest_load_with_threshold(df_sediexport, threshold):
    """Verify whether that the highest load is not accountable for more than 50 percent
    of all the load in a catchment. If this is true, adapt threshold.

    Parameters
    ----------
    df_sediexport: pandas.DataFrame
        Data Frame format of SediExport raster (format: see
        :func:`pywatemsedem.pywatemsedem.utils.raster_array_to_pandas_dataframe`)
    threshold: float
        See :func:`pywatemsedem.io.modeloutput.compute_cumulative_loads_in_sinks`

    Returns
    -------
    threshold_: float
        Adapted threshold (depending on if-else clause).
    """
    cum_sum_sinks0 = df_sediexport["cum_perc"].iloc[0]

    if cum_sum_sinks0 > threshold:
        threshold_ = np.ceil(cum_sum_sinks0)
        msg = (
            f"Highest load is accountable for more than {threshold} percent of the "
            f"total sediment load. Adapting percentage to {threshold_}."
        )
        raise Warning(msg)
    else:
        threshold_ = threshold
    return threshold_


def load_total_sediment_file(txt_total_sediment_file):
    """Load the total sediment file of CNWS written in CNWS dict_output map

    Parameters
    ----------
    txt_total_sediment_file: str or pathlib.Path | str
        File Path of the sediment file

    Returns
    -------
    dict_output: dict
        The dict_output sediment data dictionary contains the following data:

        - *erosion* (float): total amount of erosion (kg)
        - *deposition* (float): total amount of deposition (kg)
        - *river* (float): total amount of sediment run-off to the river (kg)
        - *outside_domain* (float): total amount of sediment run-off out of the \
        catchment (kg)
        - *buffers* (float): total amount of sediment trapped in buffers (kg)
        - *endpoints* (float): total sewer of sediment trapped in sewers and \
        ditches (kg)

    """
    file = open(txt_total_sediment_file, "r")
    Lines = file.readlines()

    dict_output = {}

    for line in Lines:

        line = line.split(" ")
        # hardcode
        tag = " ".join(line[:-2])
        if tag == "Total erosion:":
            dict_output["erosion"] = float(line[-2])
        elif tag == "Total deposition:":
            dict_output["deposition"] = float(line[-2])
        elif tag == "Sediment leaving the catchment, via the river:":
            dict_output["river"] = float(line[-2])
        elif tag == "Sediment leaving the catchment, not via the river:":
            dict_output["outside_domain"] = float(line[-2])
        elif tag == "Sediment trapped in buffers:":
            dict_output["buffers"] = float(line[-2])
        elif tag == "Sediment entering sewer system:":
            dict_output["endpoints"] = float(line[-2])

    return dict_output


def check_segment_edges(adj_edges, up_edges, arr):
    """Check river segment their edges. All the segments present in adj_edges
    and up_edges must be present in the array from the segments raster.

    Parameters
    ----------
    adj_edges: pandas.DataFrame
        Loaded adjacent edges text file, see :ref:`here <watemsedem:adjsegments>`
    up_edges: pandas.DataFrame
        Loaded up edges text file, see :ref:`here <watemsedem:upstrsegments>`
    arr: numpy.ndarray
        Segment raster array

    Returns
    --------
    adj_edges: pandas.DataFrame
        Filtered
    up_edges: pandas.DataFrame
        Filtered
    filt: bool
        Flag indicating filtering of dataframes
    """
    uniq_segm_rst = set(np.unique(arr))
    uniq_segm_ue = set(up_edges["upstream_line"].unique())
    uniq_segm_e = set(up_edges["line_id"].unique())
    uniq_segm_from = set(adj_edges["from"].unique())
    uniq_segm_to = set(adj_edges["to"].unique())
    filt = False

    if not uniq_segm_ue.issubset(uniq_segm_rst):
        msg = (
            "More segments in the up edges table than in the raster, filtering" "table."
        )
        warnings.warn(msg)
        up_edges = up_edges.loc[up_edges["upstream_line"].isin(uniq_segm_rst)]
        filt = True

    if not uniq_segm_e.issubset(uniq_segm_rst):
        msg = (
            "More segments in the up edges table than in the raster, filtering" "table."
        )
        warnings.warn(msg)
        up_edges = up_edges.loc[up_edges["line_id"].isin(uniq_segm_rst)]
        filt = True

    if not uniq_segm_from.issubset(uniq_segm_rst):
        msg = (
            "More segments in the adjacent edges table than in the raster, "
            "filtering table."
        )
        warnings.warn(msg)
        adj_edges = adj_edges.loc[adj_edges["from"].isin(uniq_segm_rst)]
        filt = True

    if not uniq_segm_to.issubset(uniq_segm_rst):
        msg = (
            "More segments in the adjacent edges table than in the raster, "
            "filtering table."
        )
        adj_edges = adj_edges.loc[adj_edges["to"].isin(uniq_segm_rst)]
        warnings.warn(msg)
        filt = True

    return adj_edges, up_edges, filt



import tempfile
from dataclasses import dataclass
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import rasterio

from ..defaults import ALLOWED_RASTER_FORMATS, PREFIX_TEMP
from .rasterproperties import RasterProperties
from .utils import (
    clean_up_tempfiles,
    clip_rst,
    load_raster,
    mask_array_with_val,
    set_no_data_arr,
    tiff_to_idrisi,
    write_arr_as_rst,
)


class AbstractRaster:
    """Abstract raster class based on numpy arrays and raster properties

    Parameters
    -----------
    arr: numpy.ndarray
        Raster array
    rp: RasterProperties
        See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`
    arr_mask: numpy.ndarray, default None
        See :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`
    allow_nodata_array: boolean, default False
        See :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`

    Note
    ----
    If an array mask is provided, the array is automatically masked.
    """

    def __init__(self, arr, rp, arr_mask=None, allow_nodata_array=False):
        """Initialize array and rasterproperties"""
        if len(arr.shape) < 2:
            msg = "Dimensionality of input raster array should be larger than 1."
            raise ValueError(msg)
        self._arr = arr
        self._rp = rp
        if arr_mask is not None:
            self.mask(arr_mask, allow_nodata_array)

    def _repr_html_(self):
        """notebook/ipython representation"""
        return self.arr.__array__()

    @property
    def arr(self):
        """Return array

        Returns
        -------
        numpy.ndarray
        """
        return self._arr

    @arr.setter
    def arr(self, input):
        """

        Parameters
        ----------
        input
        """
        self._arr = input

    @property
    def rp(self):
        """Return raster properties

        Returns
        -------
        :class:`pywatemsedem.geo.rasterproperties.RasterProperties`
        """
        return self._rp

    def update_nodata_value(self, to):
        """Update the nodata value

        Parameters
        ----------
        to: float
            New nodata value
        """
        self._arr[self._arr == self.rp.nodata] = to
        self._rp._nodata = to

    def clip(self):
        """NotImplemented clip function"""
        raise NotImplementedError

    def mask(self, arr_mask, allow_nodata_array=False):
        """Mask function.

        Parameters
        ----------
        arr_mask: numpy.ndarray
            Array mask (1, nodata). Note that array mask should have same nodata value
            as raster array.
        allow_nodata_array: bool
            Allow to return a nodata-array.
        """
        try:
            self._arr = set_no_data_arr(self._arr, arr_mask, self.rp.nodata)
        except ValueError:
            if allow_nodata_array:
                self._arr = self.rp.nodata * np.ones(self._arr.shape)
            else:
                self._arr = set_no_data_arr(self._arr, arr_mask, self.rp.nodata)

    def write(self, outfile_path, format="idrisi", dtype=None, nodata=None):
        """Write raster data to disk.

        Parameters
        ----------
        outfile_path: pathlib.Path or str
            File path output
        format: "idrisi" or "tiff", default "idrisi"
        dtype: numpy.dtype, default None
            Output raster type. If None, dtype of array is used.
        nodata: float, default None
            Nodata value for output raster. If None, nodata of rasterproperties is used.
        """
        outfile_path = Path(outfile_path)

        if format not in ALLOWED_RASTER_FORMATS:
            msg = (
                f"Format '{format}' not implemented in pywatemsedem. Use "
                f"{' or '.join(ALLOWED_RASTER_FORMATS)}"
            )
            raise NotImplementedError(msg)

        # check for extension and format
        if format == "idrisi":
            if outfile_path.suffix != ".rst":
                msg = (
                    f"Can not write file ('{outfile_path}')  in format 'idrisi'"
                    f" with '{outfile_path.suffix}' extension."
                )
                raise TypeError(msg)
        elif format == "tiff":
            if outfile_path.suffix != ".tif":
                msg = (
                    f"Can not write file ('{outfile_path}')  in format "
                    f"'tiff'"
                    f" with '{outfile_path.suffix}' extension."
                )
                raise TypeError(msg)

        if dtype is None:
            dtype = self._arr.dtype

        profile = self.rp.rasterio_profile.copy()
        if nodata is not None:
            profile["nodata"] = nodata

        if format == "tiff":
            write_arr_as_rst(
                self._arr,
                outfile_path,
                dtype,
                profile,
            )
        elif format == "idrisi":
            tiff_temp = Path(
                tempfile.NamedTemporaryFile(
                    suffix=".tif", prefix="pywatemsedem", delete=False
                ).name
            )
            write_arr_as_rst(self._arr, tiff_temp, dtype, profile)
            tiff_to_idrisi(tiff_temp, outfile_path, dtype=dtype)
            clean_up_tempfiles(tiff_temp, "tiff")

        return True

    def plot(self, fig=None, ax=None, nodata=None, *args, **kwargs):
        """Plot raster array with imshow

        Parameters
        ----------
        fig: matplotlib.figure.Figure, default = None
            if not given, defaults to generating new figure
        ax: matplotlib.axes.Axes, default = None
            if not given, defaults to generating new axis
        nodata: float, default = None
            Used to mask certain values present in arr which represent
            nodata (e.g. -9999)

        Returns
        -------
        fig: matplotlib.figure.Figure

        ax: matplotlib.axes.Axes
        """

        if not ax:
            fig, ax = plt.subplots(figsize=[10, 10])
        arr = mask_array_with_val(self.arr, self.arr, nodata)
        im = ax.imshow(arr, *args, **kwargs)
        plt.colorbar(im, ax=ax, orientation="vertical", shrink=0.5)

        return fig, ax

    def histogram(
        self, fig=None, ax=None, nodata=None, ylogscale=False, *args, **kwargs
    ):
        """Plot density histogram of raster data values with 25th, 50th and 75th
        percentiles

        Parameters
        ----------
        fig: matplotlib.figure.Figure, default = None
            if not given, defaults to generating new figure
        ax: matplotlib.axes.Axes, default = None
            if not given, defaults to generating new axis
        nodata: float
            Used to mask no data values
        ylogscale: bool, default = False
                Log transformation on density axis if True

        Returns
        -------
        fig: matplotlib.figure.Figure

        ax: matplotlib.axes.Axes
        """
        if not ax:
            fig, ax = plt.subplots(figsize=[10, 10])
        arr = self.arr.copy()
        if nodata is not None:
            arr = np.ma.masked_where(arr == nodata, arr)
            arr = np.ma.compressed(arr)  # fixes histogram
        arr_flat = arr.flatten()
        y, x, _ = ax.hist(arr_flat, bins="doane", density=True, *args, **kwargs)
        if ylogscale:
            ax.set_yscale("log")
        colors = ["green", "orange", "red"]
        for i, p in enumerate([25, 50, 75]):
            ax.vlines(
                np.percentile(arr.flatten(), p),
                ymin=0,
                ymax=np.max(y),
                label="p" + str(p),
                colors=colors[i],
            )
        ax.legend()
        ax.set_ylabel("density")
        return fig, ax


@dataclass
class RasterMemory(AbstractRaster):
    """Array raster

    Parameters
    ----------
    arr: numpy.ndarray
        See :class:`pywatemsedem.geo.rasters.AbstractRaster`
    rp: RasterProperties
        See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`
    arr_mask: numpy.ndarray, default None
        See :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`
    allow_nodata_array: bool, default False
        See :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`
    """

    def __init__(self, arr, rp, arr_mask=None, allow_nodata_array=False):
        """Initialize RasterMemory"""
        super().__init__(
            arr, rp, arr_mask=arr_mask, allow_nodata_array=allow_nodata_array
        )

    def clip(self):
        """NotImplemented"""
        raise NotImplementedError("Clipping not implemented for RasterMemory class")


class RasterFile(AbstractRaster):
    """Array based on a input raster file.

    Parameters
    ----------
    file_path: pathlib.Path
        File path to user input raster.
    rp: RasterProperties, default None
        See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`. If None,
        rasterproperties from inputfile are used.
    arr_mask: numpy.ndarray, default None
        See :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`
    allow_nodata_array: bool, default False
        See :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`

    Note
    ----
    If rasterproperties are inputted by the user, clipping is automatically done.
    """

    def __init__(self, file_path, rp=None, arr_mask=None, allow_nodata_array=False):
        """Initialize RasterFile class"""
        if rp:
            with rasterio.open(file_path) as src:
                rst_profile = src.profile
            if rst_profile["crs"].is_valid:
                if rst_profile["crs"] is not None:
                    if rst_profile["crs"].to_epsg() != rp.epsg:
                        msg = (
                            f"EPSG-code of {file_path} ({rst_profile['crs']}) should "
                            f"be same as epsg of input raster properties ({rp.epsg})."
                        )
                        raise IOError(msg)
            arr = self.clip(file_path, rp)
        else:
            if file_path.suffix == ".sgrd":
                file_path = file_path.with_suffix(".sdat")
            arr, profile = load_raster(file_path)
            rp = RasterProperties.from_rasterio(profile)

        super().__init__(arr, rp, arr_mask, allow_nodata_array)

    @staticmethod
    def clip(file_path, rp, resample="mode"):
        """Clip function

        Parameters
        ----------
        file_path: pathlib.Path
            File path to user input raster.
        rp: RasterProperties
            See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`
        resample: str, default "mode"
            Either "near" or "mode", see :func:`pywatemsedem.geo.utils.clip_rst`

        Returns
        -------
        arr: numpy.ndarray
            Clipped array

        Notes
        -----
        Clipping also provides resampling to another resolution. See
        :func:`pywatemsedem.geo.utils.clip_rst`.

        """
        rst_temp = tempfile.NamedTemporaryFile(
            suffix=".rst", prefix=PREFIX_TEMP, delete=False
        )
        clip_rst(
            file_path,
            Path(rst_temp.name),
            rp.gdal_profile,
            resampling=resample,
        )
        arr, profile = load_raster(Path(rst_temp.name))
        shape = arr[arr != profile["nodata"]].shape
        rst_temp.close()
        clean_up_tempfiles(rst_temp, "rst")

        if shape == (0,):
            msg = (
                f"Clipped output raster is empty. Make sure your input raster covers "
                f"your defined spatial extent (bounds: {rp.bounds}, resolution:"
                f" {rp.resolution}, espg: {rp.epsg})."
            )
            raise IOError(msg)

        return arr


class TemporalRaster:
    """3-D raster with first two dimension spatial x and y dimension, with the third
    dimension being temporal.

    Parameters
    ----------
    arr: numpy.ndarray
        3D Raster array
    rp: RasterProperties
        See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`
    arr_mask: np.ndarray, default None
        See :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`

    Note
    ----
    This class only works with array inputs!
    """

    def __init__(self, arr, rp, arr_mask=None):
        """Initialize array and rasterproperties"""
        if len(arr.shape) != 3:
            msg = "Dimensionality of temporal raster array should be equal to 3."
            raise ValueError(msg)
        self._arr = arr
        self._rp = rp
        self._arr_mask = arr_mask

    @property
    def arr(self):
        """Return array
        Returns
        -------
        numpy.ndarray
        """
        return self._arr

    def write(self, outfiles, format="idrisi", dtype=None):
        """Write function.

        Parameters
        ----------
        outfiles: list or tuple of pathlib.Path or str
            List of output files
        format: basestring
            See :func:`pywatemsedem.geo.rasters.AbstractRaster.write`
        dtype: numpy.dtype
            See :func:`pywatemsedem.geo.rasters.AbstractRaster.write`

        """
        if len(outfiles) != self._arr.shape[-1]:
            msg = (
                f"Number of outputfiles ('{outfiles}') should be equal to number of "
                f"arrays in the third dimension('{self._arr.shape[2]}')."
            )
            raise ValueError(msg)

        for i in range(self._arr.shape[-1]):
            rm = RasterMemory(self._arr[:, :, i], self._rp, arr_mask=self._arr_mask)
            rm.write(outfiles[i], format, dtype)

        return True

    def plot(self, nodata=None, *args, **kwargs):
        """Sequential plot raster in cols

        Parameters
        ----------
        nodata: float
        """
        shape = self._arr.shape[-1]
        fig, ax = plt.subplots(ncols=shape, figsize=[10, 5 * shape])
        for i in range(shape):
            rm = RasterMemory(self._arr[:, :, i], self._rp, arr_mask=self._arr_mask)
            ax_i = ax[i] if shape > 1 else ax
            rm.plot(ax_i, nodata=nodata, *args, **kwargs)



import functools

import matplotlib.colors as mplcolors
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np

from ..geo.utils import mask_array_with_val


def plot_continuous_raster(
    fig,
    ax,
    arr,
    bounds,
    cmap="cividis",
    norm=None,
    colorbar=True,
    ticks=None,
    *args,
    **kwargs
):
    """
    Parameters
    ----------
    fig: matplotlib.figure.Figure

    ax: matplotlib.pyplot.axis

    arr: numpy.ndarray
        Contains raster data
    bounds: list
        [left, bottom, right, upper] as obtained from RasterProperties.bounds
        (See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`)
    cmap: string or matplotlib.colors.Colormap, default = 'cividis'

    norm: matplotlib.colors.Normalize, optional
        normalisation function for colormapping
    colorbar: bool, default = True
                Choice of displaying a colorbar
    ticks: list, optional
        Supply [min, 25th percentile, 50th percentile ,75th percentile, max]
        when norm is used!

    Returns
    -------
    fig: matplotlib.figure.Figure

    ax: matplotlib.pyplot.axis

    """
    im = ax.imshow(
        arr,
        cmap=cmap,
        norm=norm,
        extent=np.array(bounds)[[0, 2, 1, 3]],
        *args,
        **kwargs,
        interpolation="nearest",
    )
    if colorbar:
        fig.colorbar(im, orientation="vertical", shrink=0.5, ticks=ticks)
    else:  # allows custom colorbar
        fig = im
    return fig, ax


def plot_discrete_raster(
    fig, ax, arr, bounds, labels, cmap="cividis", norm=None, *args, **kwargs
):
    """
    Parameters
    ----------
    fig: matplotlib.figure.Figure

    ax: matplotlib.pyplot.axis

    arr: numpy.ndarray
        Contains raster data
    bounds: list
        [left, bottom, right, upper] as obtained from RasterProperties.bounds
        (See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`)
    labels: list
        Supply unique string for each value
    cmap: string or matplotlib.colors.Colormap, default = 'cividis

    norm: matplotlib.colors.Normalize, optional
        normalisation function for colormapping
    Returns
    -------
    fig: matplotlib.figure.Figure

    ax: matplotlib.pyplot.axis

    """
    im = ax.imshow(
        arr,
        cmap=cmap,
        norm=norm,
        extent=np.array(bounds)[[0, 2, 1, 3]],
        interpolation="nearest",
        *args,
        **kwargs,
    )
    values = np.unique(arr)
    colours = [im.cmap(im.norm(value)) for value in values]
    # create a patch (proxy artist) for every color
    patches = [
        mpatches.Patch(
            color=colours[i],
            label=labels[
                np.where(values[i] == np.arange(np.min(values), np.max(values) + 1))[0][
                    0
                ]
            ],
        )
        for i in range(len(values) - 1)
    ]
    ax.legend(handles=patches, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)
    return fig, ax


def plot_output_raster(
    fig, ax, arr, mask, bounds, title="none", ticks=None, *args, **kwargs
):
    """Standard non-linear continuous raster plot for output raster

    Parameters
    ----------
    fig: matplotlib.figure.Figure

    ax: matplotlib.pyplot.axis

    arr: numpy.ndarray
        Contains raster data
    mask: numpy.ndarray
        Contains raster data, 0 outside of domain
    bounds: list
        [left, bottom, right, upper] as obtained from RasterProperties.bounds
        (See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`)
    cmap: string or matplotlib.colors.Colormap, default = 'cividis'

    ticks: list, optional
        When no list is given, 0th, 25th, 50th, 75th and 100th percentile are used

    Returns
    -------
    fig: matplotlib.figure.Figure

    ax: matplotlib.pyplot.axis
    """
    fig, ax = axes_creator(fig, ax)
    arr = mask_array_with_val(arr, mask, 0)
    arr = np.ma.filled(arr, np.nan)
    if not ticks:
        ticks = np.nanpercentile(arr, [0, 25, 50, 75, 100])
    norm = normalising_function(ticks[0], ticks[1], ticks[2], ticks[3], ticks[4])
    fig, ax = plot_continuous_raster(
        fig,
        ax,
        arr=arr,
        bounds=bounds,
        norm=norm,
        ticks=ticks,
        *args,
        **kwargs,
    )
    ax.set_title(title)
    return fig, ax


def log_scale_enabler(arr, cnorm):
    """Checks if log scale is supplied.
    If minimum is lower or equal to zero,
    changes lower bound to smallest value above zero.

    Parameters
    ----------
    arr: numpy.ndarray
        rastervalues
    cnorm: str
        cnorm used for hvplot

    Returns
    -------
    lower: float
        lower value for colorscale
    """
    if cnorm == "log":
        if np.nanmin(arr) <= 0:
            lower = np.nanmin(arr[arr > 0])
        else:
            lower = np.nanmin(arr)
    else:
        lower = np.nanmin(arr)
    return lower


def axes_creator(fig, ax):
    """Creates fig and ax object if none ar given

    Parameters
    ----------
    fig: matplotlib.figure.Figure

    ax: matplotlib.pyplot.axis

    Returns
    --------
    fig: matplotlib.figure.Figure

    ax: matplotlib.pyplot.axis

    """
    if not ax:
        fig, ax = plt.subplots(figsize=[10, 10])
    return fig, ax


def normalising_function(mini, Q25, Q50, Q75, maxi):
    """Custom normalisation function for maptlotlib plotting

    Parameters
    ----------
    mini
        minimum of the colorscale
    Q25
        25th percentile
    Q50
        50th percentile
    Q75
        75th percentile
    maxi
        maximum of the colorscale

    Returns
    -------
    norm: matplotlib.colors.Normalize


    """
    _forward_custom = functools.partial(
        _forward, mini=mini, Q25=Q25, Q50=Q50, Q75=Q75, maxi=maxi
    )
    _inverse_custom = functools.partial(
        _inverse, mini=mini, Q25=Q25, Q50=Q50, Q75=Q75, maxi=maxi
    )
    norm = mplcolors.FuncNorm((_forward_custom, _inverse_custom), vmin=mini, vmax=maxi)
    return norm


def _forward(x, mini, Q25, Q50, Q75, maxi):
    """Custom forward normalisation function
    performs uniform interpolation between percentiles

    Parameters
    ----------
    x: numpy.ndarray

    mini
        minimum of the colorscale
    Q25
        25th percentile
    Q50
        50th percentile
    Q75
        75th percentile
    maxi
        maximum of the colorscale

    Returns
    -------
    y: numpy.ndarray

    """
    y = x.copy()
    y = y * np.nan
    cond = (x <= Q25) * (x >= mini)
    y[cond] = 0.25 / (Q25 - mini) * (x[cond] - mini)
    cond2 = (x > Q25) * (x <= Q50)
    y[cond2] = (0.5 - 0.25) / (Q50 - Q25) * (x[cond2] - Q25) + 0.25
    cond3 = (x <= Q75) * (x > Q50)
    y[cond3] = (0.75 - 0.5) / (Q75 - Q50) * (x[cond3] - Q50) + 0.5
    cond4 = (x <= maxi) * (x > Q75)
    y[cond4] = (1 - 0.75) / (maxi - Q75) * (x[cond4] - Q75) + 0.75
    return y


def _inverse(y, mini, Q25, Q50, Q75, maxi):
    """Custom inverse normalisation function
    performs uniform interpolation between percentiles

    Parameters
    ----------
    x: numpy.ndarray

    min
        minimum of the colorscale
    Q25
        25th percentile
    Q50
        50th percentile
    Q75
        75th percentile
    max
        maximum of the colorscale

    Returns
    -------
    y: numpy.ndarray
    """

    x = y.copy()
    x = x * np.nan
    cond = (y >= 0) * (y <= 0.25)
    x[cond] = (Q25 - mini) / 0.25 * y[cond] + mini
    cond2 = (y >= 0.25) * (y <= 0.5)
    x[cond2] = (Q50 - Q25) / 0.25 * (y[cond2] - 0.25) + Q25
    cond3 = (y <= 0.75) * (y > 0.5)
    x[cond3] = (Q75 - Q50) / 0.25 * (y[cond3] - 0.5) + Q50
    cond4 = (y <= 1) * (y > 0.75)
    x[cond4] = (maxi - Q75) / 0.25 * (y[cond4] - 0.75) + Q75
    return x



class InputFileName:
    """Holds template inputfile names for WaTEM/SEDEM input

    Parameters
    ----------
    format: str | pathlib.Path
        Database format.
    """

    def __init__(self, format="filebased"):

        if format == "filebased":
            self.kfactor_file = "kfactor.rst"
            self.dtm_file = "dtm.rst"
            self.pfactor_file = "pfactor.rst"
            self.segments_file = "segments.rst"
            self.routing_file = "routing.rst"
            self.parcelmosaic_file = "parcels_landuse.rst"
            self.cfactor_file = "cfactor.rst"
            self.outlet_file = "outlet.rst"
            self.ktc_file = "ktc_map.rst"
            self.mask_file = "mask.rst"
            self.ditches_file = "ditches.rst"
            self.conductivedams_file = "conductivedams.rst"
            self.endpoints_file = "sewers.rst"
            self.endpoints_id_file = "sewers_id.rst"
            self.buffers_file = "buffers.rst"
            self.cn_file = "cn.rst"
            self.adjacentedges_file = "adjacent_edges.txt"
            self.upedges_file = "up_edges.txt"

        else:
            msg = f"Database format '{format} not implemented."
            raise NotImplementedError(msg)



"""Extra functions that can be used for plotting"""

import logging
import sys

# GIS and datahandling libraries
try:
    import matplotlib.pyplot as plt
    import pandas as pd
except ImportError as e:
    logging.error("not all necessary libraries are available for import!")
    logging.error(e)
    sys.exit()

logger = logging.getLogger(__name__)


def plot_cumulative_sedimentload(df, fname):
    """Make a cumulative plot of the ordered sediment load values (high to low)

    Parameters
    ----------
    df: pandas.DataFrame
        Data with cumulative sediment load
        - *value* (float): sediment load value
        - *cum_perc* (float): cumulative percentage
        - *rank* (float): rank
    percentage: int
        Percentage (%) of cumulative sediment load that is to be analysed
    fname: str or pathlib Path
        File path of output figure
    """

    fig, ax = plt.subplots(2, 1, figsize=[10, 7.5])

    x = df.loc[:, "value"]
    y = df.loc[:, "cdf"]
    ax[0].plot(
        x, y, color=[0.2] * 3, label=r"Cumulative percentage of sediment load (%)"
    )
    ax[1].plot(
        df.loc[:, "rank"],
        y,
        color=[0.2] * 3,
        label=r"Cumulative percentage of sediment load (%)",
    )
    ax[0].set_xlabel("Sediment load (-)")
    ax[0].set_ylabel("CDF (%)")
    ax[1].set_xlabel("Ranked id (-)")
    ax[1].set_ylabel("CDF (%)")
    ax[1].legend()
    plt.savefig(fname, bbox_inches="tight")
    plt.close()


def plot_time_series_for_in_river_points(
    variable, folder, resmap="", convert_output=True, output_per_segment=False
):
    """
    Plot timeseries. All timeseries are stored in the postprocesfolder as a png-file.

    Parameters
    ----------
    variable: string
        'Q': plot with discharge and rainfall vs time
        'Sedigram': plot with sediment concentration vs time
        'Sediment': plot with total sediment mass vs time
    folder: pathlib Path, optional
        path to folder where CNWS results are saved
    resmap: pathlib Path, optional
        path to folder to save pngs
    convert_output: bool, optional
        True if units in minutes
    output_per_segment: bool, optional
        True if plot has to be per segment
    """
    if variable == "Q":
        base_name = "Discharge"  # TO DO: ook neerslag!
        baseTitle = "Debiet in "
        ylable = "Debiet [m^3/s]"
    elif variable == "Sedigram":
        base_name = "Sediment concentration"
        baseTitle = "Sedigram "
        ylable = "Sediment concentratie [g/l]"
    elif variable == "Sediment":
        base_name = "Sediment"
        baseTitle = "Sediment aanvoer in "
        ylable = "Sedimentvracht [kg]"
    else:
        msg = "No valable type defined for plotting timeseries"
        logger.warning(msg)
        return

    if convert_output:
        indexcol = "Time (min)"
    else:
        indexcol = "Time (sec)"

    txt_out = folder / f"{base_name}.txt"
    if txt_out.exists():
        df_out = pd.read_csv(txt_out, sep="\t", header=1)
        df_out.index = df_out[indexcol]
        df_out = df_out.drop(indexcol, 1)
    else:
        msg = f"{base_name}.txt not found in modelresults!"
        raise logging.error(msg)

    if output_per_segment:
        txt_segm = folder / f"{base_name}_VHA.txt"
        if txt_segm.exists():
            df_segm = pd.read_csv(txt_segm, sep="\t", header=1)
            df_segm.index = df_segm[indexcol]
            df_segm = df_segm.drop(indexcol, 1)
        else:
            msg = f"{base_name}_VHA.txt not found in modelresults!"
            raise logging.error(msg)
        df = pd.concat([df_out, df_segm], axis=1, join="inner")
    else:
        df = df_out.copy()

    for col in df.columns:
        if col.startswith("Unnamed"):
            df = df.drop(col, 1)

    for i, col in enumerate(df.columns):
        plt.figure(i)
        plt.ylabel(ylable)
        if col.startswith("VHA "):
            title = baseTitle + f"segment {col[12:]}"
        else:
            title = baseTitle + f"outlet {col[7:]}"
        plt.title(title)
        df[col].plot()
        plt.savefig(resmap / title + ".png")
        plt.close(i)



# implicit testing of the modelinput class

from pathlib import Path

import pytest

from pywatemsedem.io.modelinput import Modelinput

# testing normal functioning


@pytest.mark.skip(
    reason="test gives saga-issues in conda pipeline, ommited, raised as"
    " issue in saga"
)
def test_modelinput_all():
    """tests for Modelinput class"""
    # folder = Path(r"pywatemsedem/tests/io/data")
    folder = Path("tests") / "io" / "data"
    filepath = folder / "model_in"
    P_ex = filepath / "pfactor.rst"
    example = Modelinput(P_ex, resolution=20, epsg=31370, nodata=-9999)
    # C-factor
    example.cfactor = filepath / "cfactor.rst"
    example.cfactor.plot()
    # example.cfactor.hv_plot()
    # Buffers: no data for langegracht
    example.buffers = filepath / "buffers.rst"
    example.buffers.plot()
    # example.buffers.hv_plot()
    # DTM
    example.dtm = filepath / "dtm.rst"
    example.dtm.plot()
    # example.dtm.hv_plot()
    # K-factor
    example.kfactor = filepath / "Kfactor.rst"
    example.kfactor.plot()
    # example.kfactor.hv_plot()
    # kTC
    example.ktc = filepath / "ktc.rst"
    example.ktc.plot()
    # example.ktc.hv_plot()
    # Outlet
    example.outlet = filepath / "Outlet.rst"
    example.outlet.plot()
    # example.outlet.hv_plot()
    # P-factor
    example.pfactor = P_ex
    example.pfactor.plot()
    # example.pfactor.hv_plot()
    # landuseparcels
    example.compositelanduse = filepath / "perceelskaart.rst"
    # example.landuseparcels.plot()
    # example.landuseparcels.hv_plot()
    # PTEF
    example.ptef = filepath / "PTEFmap.rst"
    example.ptef.plot()
    # example.ptef.hv_plot()
    # riversegments
    example.riversegments = filepath / "segments.rst"
    example.riversegments.plot()
    # example.riversegments.hv_plot()
    # riverrouting
    example.riverrouting = filepath / "routing.rst"
    example.riverrouting.plot()
    # example.riverrouting.hv_plot()
    # sewers
    example.sewers = filepath / "sewers.rst"
    example.sewers.plot()
    # example.sewers.hv_plot()
    # upstream segments
    example.upstreamsegments = filepath / "upedges.txt"
    # adjecant segments
    example.adjacentsegments = filepath / "adjedges.txt"



import pytest
from pyproj import CRS
from pyproj.exceptions import CRSError
from rasterio import Affine

from pywatemsedem.geo.rasterproperties import RasterProperties


def test_rasterproperties():
    """Test roundtrip between gdal and rasterio raster profile"""
    rasterio_profile = {
        "driver": "GTiff",
        "nodata": -9999.0,
        "width": 294,
        "height": 509,
        "count": 1,
        "crs": CRS.from_epsg(31370),
        "transform": Affine(20.0, 0.0, 201620.0, 0.0, -20.0, 164060.0),
        "compress": "deflate",
    }
    gdal_profile = {
        "nodata": -9999.0,
        "epsg": "EPSG:31370",
        "res": 20.0,
        "minmax": [201620.0, 153880.0, 207500.0, 164060.0],
        "ncols": 294,
        "nrows": 509,
    }

    # compare gdal profile
    rp1 = RasterProperties.from_rasterio(rasterio_profile)
    assert rp1.gdal_profile == gdal_profile

    # compare rasterio profile
    rp2 = RasterProperties.from_gdal(gdal_profile)
    assert rp2.rasterio_profile == rasterio_profile

    # test from bounds
    bounds = [201620.0, 153880.0, 207500.0, 164060.0]
    resolution = 20
    nodata = -9999
    epsg = 31370
    rp4 = RasterProperties(bounds, resolution, nodata, epsg)
    assert rp4.rasterio_profile == rasterio_profile

    # test with incomplete rasterio profile
    rasterio_profile_incompl = {
        "driver": "GTiff",
        "nodata": -9999.0,
        "width": 294,
        "height": 509,
        "count": 1,
        "transform": Affine(20.0, 0.0, 201620.0, 0.0, -20.0, 164060.0),
    }

    with pytest.raises(IOError) as excinfo:
        RasterProperties.from_rasterio(rasterio_profile_incompl)
    assert "Function input is not a rasterio profile instance!" in str(excinfo.value)

    # test with incomplete gdal profile
    gdal_profile_incompl = {
        "nodata": -9999.0,
        "epsg": "EPSG:31370",
        "res": 20.0,
        "ncols": 294,
        "nrows": 509,
    }
    with pytest.raises(IOError) as excinfo:
        RasterProperties.from_gdal(gdal_profile_incompl)
    assert "Function input is not a gdal profile instance!" in str(excinfo.value)

    # test unknown epsg
    epsg = -11000
    with pytest.raises(CRSError) as excinfo:
        RasterProperties(bounds, resolution, nodata, epsg)
    assert "is an unknown " in str(excinfo.value)

    # test wrong format epsg
    epsg = "-11000"
    with pytest.raises(TypeError) as excinfo:
        RasterProperties(bounds, resolution, nodata, epsg)
    assert "need to be an integer code." in str(excinfo.value)

    # test not supported driver
    epsg = 31370  # change epsg back to valid format!

    with pytest.raises(IOError) as excinfo:
        RasterProperties(bounds, resolution, nodata, epsg, driver="tsjaarbomb32")
    assert "Raster property driver" in str(excinfo.value)



from functools import wraps
from inspect import signature
from pathlib import Path

import fiona
import numpy as np
import rasterio
from fiona.errors import DriverError
from rasterio.errors import RasterioIOError

from .rasterproperties import RasterProperties


class PywatemsedemInputError(Exception):
    """Raise when input data are not conform the pywatemsedem required input format."""


class PywatemsedemTypeError(Exception):
    """Raise when input data type are not conform the pywatemsedem required type format."""


def valid_polygonvector(vct, fun):
    """Check if input vector file is a valid polygon shape

    Parameters
    ----------
    vct: pathlib.Path
        File path to vector.
    fun: callable
        See :func:`pywatemsedem.geo.valid.valid_input`.
    """
    valid_exists(vct, fun)
    valid_vector(vct, fun, "Polygon")

    return True


def valid_pointvector(vct, fun):
    """Check if input vector file is a valid point shape

    Parameters
    ----------
    vct: pathlib.Path
        File path to vector.
    fun: callable
        See :func:`pywatemsedem.geo.valid.valid_input`.
    """
    valid_exists(vct, fun)
    valid_vector(vct, fun, "Point")

    return True


def valid_linesvector(vct, fun):
    """Check if input vector file is a valid lines shape

    Parameters
    ----------
    vct: pathlib.Path
        File path to vector.
    fun: callable
        See :func:`pywatemsedem.geo.valid.valid_input`.
    """
    valid_exists(vct, fun)
    valid_vector(vct, fun, "LineString")

    return True


def valid_rasterlist(lst_rst, fun):
    """Check if input is a valid list of rasters

    Parameters
    ----------
    lst_rst: list
        List of file paths to rasters.
    fun: callable
        See :func:`pywatemsedem.geo.valid.valid_input`.
    """
    if type(lst_rst) is not list:
        msg = (
            f"Input '{lst_rst}' is not a valid list of rasters input, "
            f"cannot execute '{fun.__name__}'."
        )
        raise PywatemsedemInputError(msg)

    for rst in lst_rst:
        valid_raster(rst, fun)

    return True


def valid_vectorlist(lst_vct, fun, req_type=None):
    """Check if input is a valid list of rasters

    Parameters
    ----------
    lst_vct: list
        List of file paths to rasters.
    fun: callable
        See :func:`pywatemsedem.geo.valid.valid_input`.
    req_type:
        Required geometry type of vector. See :func:`pywatemsedem.geo.valid.valid_vector`

    Note
    -----
    req_type can only be one type, not a mix.
    """
    if type(lst_vct) is not list:
        msg = (
            f"Input '{lst_vct}' is not a valid list of vectors input, "
            f"cannot execute '{fun.__name__}'."
        )
        raise PywatemsedemInputError(msg)

    for vct in lst_vct:
        valid_vector(vct, fun, req_type)

    return True


def valid_vector(vct, fun, req_type=None):
    """Check if input file is a valid vector

    Parameters
    ----------
    Parameters
    ----------
    vct: pathlib.Path
        File path to vector.
    fun: callable
        See :func:`pywatemsedem.geo.valid.valid_input`.
    req_type:
        Required geometry type of vector, limited to "Polygon", "LineString", "Point"
        and None (i.e. don't check).
    """
    if req_type not in ["Polygon", "LineString", "Point", None]:
        msg = f"Geomtry type '{req_type}' not known."
        raise IOError(msg)
    try:
        c = fiona.open(vct)
        if req_type is not None:
            if c.schema["geometry"] is not req_type:
                msg = (
                    f"Geometry input type ('{c.schema['geometry']}') of '{vct}' is "
                    f"not of "
                    f"type '{req_type}', cannot execute '{fun.__name__}'."
                )
                raise PywatemsedemTypeError(msg)
        c.close()

    except DriverError:
        msg = (
            f"The fiona engine in pywatemsedem cannot open '{vct}' as it is not "
            f"recognized as a supported file format, cannot execute '{fun.__name__}'."
        )
        raise PywatemsedemTypeError(msg)

    return True


def valid_raster(rst, fun):
    """Check if input file is a valid raster

    Parameters
    ----------
    rst: pathlib.Path
        File path to raster.
    fun: callable
        See :func:`pywatemsedem.geo.valid.valid_input`.
    """

    valid_exists(rst, fun)

    try:
        c = rasterio.open(rst)
        c.close()
    except RasterioIOError:
        msg = (
            f"The rasterio engine in pywatemsedem cannot open '{rst}' as it is not "
            f"recognized as a supported file format, cannot execute '{fun.__name__}'."
        )
        raise PywatemsedemTypeError(msg)

    return True


def valid_exists(rst, fun):
    """Check if input file exists

    Parameters
    ----------
    rst: pathlib.Path
        File path to raster.
    fun: callable
        See :func:`pywatemsedem.geo.valid.valid_input`.
    """

    if not Path(rst).exists():
        if fun is not None:
            msg = (
                f"Input raster '{rst}' does not exist, cannot execute "
                f"'{fun.__name__}'."
            )
        else:
            msg = f"Input raster '{rst}' does not exist"
        raise PywatemsedemInputError(msg)

    return True


VALID_FUN = [
    "valid_vectorlist",
    "valid_pointvector",
    "valid_polygonvector",
    "valid_linesvector",
    "valid_raster",
    "valid_vector",
    "valid_rasterlist",
]


def valid_input(func=None, dict=None):
    """Customizable wrapper function that allows to check arg-defined function input.

    This wrapper is defined to check formats of file-based raster and vector input. It
    makes use of a dictionary to apply specific valid functions on non-keyword
    arguments.  This valid function is typically applied to functions which have file
    paths that have to be checked in their input.

    Parameters
    ----------
    func: callable
        To call function for which to check inputs. Note that only function with all
        keyword-arguments as parameters can be used.
    dict: dictionary
        Holding string of fun parameters as keys, and valid-callable function as
        values.

    Returns
    -------
    function output

    Examples
    --------
    Use with a '@'-decorator to check input types for utils functions:

    ::

        from pywatemsedem.geo.utils import valid_input,valid_rasterlist,valid_polygonvector
        #note: only on-keyword arguments!
        @valid_input(dict={"lst_rst": valid_rasterlist, "vct_in": valid_polygonvector})
        def grid_statistics(lst_rst,vct_in):
        # ... function code

    Above example check if input of lst_rst is a list of rasters (and rasters exist)
    (1) and if input of vct_in is a polygoon shape (2).

    Note
    -----
    1. Can only be applied to non-keyword arguments.
    2. Validation methods should be added to VALID_FUN.
    """
    assert callable(func) or func is None

    def _decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for it, key in enumerate(signature(func).parameters.keys()):
                if key in dict:
                    valid_fun = dict[key]
                    if valid_fun.__name__ not in VALID_FUN:
                        msg = (
                            f"Validation fuction {valid_fun} for {key} not "
                            f"implemented"
                        )
                        raise KeyError(msg)
                    else:
                        valid_fun(args[it], func)
            return func(*args, **kwargs)

        return wrapper

    return _decorator(func) if callable(func) else _decorator


def valid_rp(rp):
    """Check if rasterproperties is not equal to none and valid

    Parameters
    ----------
    rp: pywatemsedem.geo.rasterproperties.RasterProperties
    """
    if rp is None:
        msg = "Please define rasterproperties"
        raise PywatemsedemInputError(msg)
    if not isinstance(rp, RasterProperties):
        msg = (
            f"Input RasterProperties '{rp}' is not a valid instance of the "
            f"'{RasterProperties}' class"
        )
        raise PywatemsedemTypeError(msg)

    return True


def valid_mask(mask):
    """Check if mask is not equal to none and valid

    Parameters
    ----------
    mask
    """
    if mask is None:
        msg = "Please define a mask"
        raise PywatemsedemInputError(msg)
    if set(np.unique(mask)) != {0, 1}:
        msg = "Please define a valid mask with only 0,1-values"
        raise PywatemsedemTypeError(msg)

    return True



import pytest

from pywatemsedem.valid import valid_req_property


class ExampleClass:
    """Example class for testing decorator pywatemsedem.valid.valid_req_property"""

    _req_property = None

    @property
    def non_mandatory(self):
        """Define a property based on a non-mandatory required property"""
        valid_req_property(
            self,
            current_property="Non mandatory property",
            req_property_name="req_property",
            mandatory=False,
        )
        return self._req_property + 5

    @property
    def mandatory(self):
        """Define a property based on a mandatory required property"""
        valid_req_property(
            self,
            current_property="mandatory property",
            req_property_name="req_property",
            mandatory=True,
        )
        return self._req_property + 5

    @property
    def req_property(self):
        """Getter req propery"""
        return self._req_property

    @req_property.setter
    def req_property(self, value):
        """Setter req propery with float value"""
        self._req_property = value


class TestValidSourceProperty:
    """This test class is used to check functioning of the required property decorator
    pywatemsedem.valid.valid_req_property

    The required property decorator checks whether a required property is properly
    defined. A distinction is made between mandatory or non-mandatory properties:

    - *mandatory*: The class needs the required property X to be defined (not equal to
      None) when calling property Y, otherwise it raises an Error.
    - *non_mandatory*: The class needs the required property X to be defined (not equal
      to None) when calling property Y, otherwise it returns a None.
    """

    def test_non_mandatory(self):
        """Test for non mandatory required property"""
        source = ExampleClass()
        source.req_property = 8

        assert source.non_mandatory == 13

    def test_mandatory(self):
        """Test for mandatory required property"""
        source = ExampleClass()
        err_msg = "Please first set mandatory property"
        with pytest.raises(IOError, match=err_msg):
            print(source.mandatory)

        source.req_property = 3
        assert source.non_mandatory == 8



import tempfile
from pathlib import Path

import geopandas as gpd
import numpy as np

from ..defaults import PREFIX_TEMP
from .rasterproperties import RasterProperties
from .utils import (
    clean_up_tempfiles,
    clip_vct,
    create_spatial_index,
    get_geometry_type,
    lines_to_direction,
    lines_to_raster,
    load_raster,
    polygons_to_raster,
    read_rasterio_profile,
    vct_to_rst_field,
)


class AbstractVector:
    """Abstract Vector class based on geopandas GeoDataFrame"""

    def __init__(
        self, geodata, geometry_type, req_geometry_type=None, allow_empty=False
    ):
        """Abstract vector class

        Parameters
        ----------
        geodata: geopandas.GeoDataFrame
            Input data set.
        geometry_type: str
            Geometry type of input dataset
        req_geometry_type: str, default None
            Geometry type, for implemented types, see
            :func:`pywatemsedem.geo.vectors.AbstractVector.check_type`
        allow_empty: bool, default False
            Allow an empty geodataframe
        """
        self._geodata = geodata
        self._geometry_type = geometry_type
        self.check_type(geometry_type, req_geometry_type)
        if not allow_empty:
            try:
                self.check_if_empty()
            except ValueError:
                msg = (
                    "Input vector is empty. If you wish to "
                    "return an empty vector, please use 'allow_empty'."
                )
                raise ValueError(msg)

    def check_type(
        self,
        geometry_type,
        req_geometry,
        implemented_types=["LineString", "Polygon", "Point"],
    ):
        """Check geometry types of vector to the required type

        Parameters
        ----------
        geometry_type: str
            Geometry type of input dataset.
        req_geometry: str
            The required geometry types
        implemented_types: list, default "LineString", "Polygon", "Point"
            List of implemented geometry types.
        """
        if req_geometry is not None:
            if req_geometry not in implemented_types:
                msg = (
                    f"Required geometry item type '{req_geometry}' not known to "
                    f"pywatemsedem. Please select '{' or '.join(implemented_types)}'"
                )
                raise TypeError(msg)
            if req_geometry is not geometry_type:
                msg = (
                    f"Input vector should have geometry item type "
                    f"'{req_geometry}', not '{geometry_type}'."
                )
                raise TypeError(msg)

    def check_if_empty(self):
        """Check if input is empty"""
        if self._geodata.empty:
            msg = "Input vector cannot be empty!"
            raise ValueError(msg)

    def plot(self, color=None, column=None):
        """Plot shape vector with geopandas plot

        Parameters
        ----------
        color

        Returns
        -------
        ax: matplotlib.pyplot.axis
        """
        df = self.geodata.to_crs(epsg=3857)
        kwargs = {}
        if column is not None:
            kwargs = {**kwargs, "cmap": "viridis", "legend": True}
        ax = df.plot(
            alpha=0.8, lw=2, figsize=[10, 10], color=color, column=column, **kwargs
        )

        return ax

    @property
    def geodata(self):
        """Property to override"""
        return self._geodata

    @geodata.setter
    def geodata(self, input):
        """Setter"""
        self._geodata = input

    def clip(self):
        """NotImplemented clip function"""
        raise NotImplementedError

    def write(self, outfile_path):
        """Write raster data to disk.

        Parameters
        ----------
        outfile_path: pathlib.Path or str, default None
            File path output
        """
        if outfile_path is not None:
            outfile_path = Path(outfile_path)
        if outfile_path.suffix == ".shp":
            self._geodata.to_file(outfile_path)
            create_spatial_index(outfile_path)
        elif outfile_path.suffix == "":
            self._geodata.to_file(outfile_path, driver="ESRI Shapefile")  # esri
            create_spatial_index(outfile_path)
        else:
            msg = f"Extension of  {outfile_path} not support in pywatemsedem"
            raise TypeError(msg)
        return True

    def rasterize(
        self,
        rst_reference,
        epsg,
        col="NR",
        nodata=None,
        dtype_raster="float",
        convert_lines_to_direction=False,
        gdal=False,
    ):
        """Rasterize function for shape file

        Parameters
        ----------
        rst_reference: str or  pathlib.Path
            File path to reference file for raster output.
        epsg: int
            EPSG code should be a numeric value, see https://epsg.io/.
        col: str, default "NR"
            Column name to map
        nodata: float, default None
            Values within dataframe 'col' that have to be considered as nodata in
            raster.
        dtype_raster: str, default "float"
            Output raster type
            convert_lines_to_direction:
        convert_lines_to_direction: bool, default "False"
            Convert lines to directions
        gdal: bool, default False
            Use gdal(true) / saga (false)-enige for mapping.

        Returns
        -------
        arr: numpy.ndarray
            Return numpy array
        """
        # convert lines to directions only be done with saga
        if gdal & convert_lines_to_direction:
            gdal = False

        if (col == "NR") & ("NR" not in self._geodata.columns):
            self._geodata["NR"] = np.arange(0, len(self._geodata), 1)
        if nodata is not None:
            self._geodata.loc[self._geodata[col] == nodata, col] = -99999.0
        vct_temp = tempfile.NamedTemporaryFile(prefix=PREFIX_TEMP, suffix=".shp").name
        self.write(vct_temp)

        rp = RasterProperties.from_rasterio(
            read_rasterio_profile(rst_reference), epsg=epsg
        )
        if gdal:
            tf_rst = tempfile.NamedTemporaryFile(
                suffix=".tif", prefix=PREFIX_TEMP, delete=False, mode="w"
            )
            vct_to_rst_field(vct_temp, tf_rst.name, rp.gdal_profile, col)
            arr, _ = load_raster(tf_rst.name)
            clean_up_tempfiles(tf_rst, "tiff")

        else:
            tf_rst = tempfile.NamedTemporaryFile(
                suffix=".sgrid", prefix=PREFIX_TEMP, delete=False
            )
            if self._geometry_type == "LineString":
                if convert_lines_to_direction:
                    lines_to_direction(vct_temp, Path(tf_rst.name), rst_reference)
                else:
                    lines_to_raster(
                        vct_temp, Path(tf_rst.name), rst_reference, col, dtype_raster
                    )
            elif self._geometry_type == "Polygon":
                if convert_lines_to_direction:
                    msg = "Cannot convert polygons to directions"
                    raise IOError(msg)
                polygons_to_raster(
                    vct_temp, Path(tf_rst.name), rst_reference, col, dtype_raster
                )
            elif self._geometry_type == "Point":
                msg = "Rasterisation of points is not implemented."
                raise NotImplementedError(msg)

            arr, profile = load_raster(Path(tf_rst.name).with_suffix(".sdat"))
            # correct no data value if necessary
            if profile["nodata"] != rp.nodata:
                arr[arr == profile["nodata"]] = rp.nodata
            tf_rst.close()
            clean_up_tempfiles(tf_rst, "tiff")
            clean_up_tempfiles(tf_rst, "saga")

        clean_up_tempfiles(vct_temp, "shp")

        return arr


class VectorMemory(AbstractVector):
    """Geopandas vector

    Parameters
    ----------
    geodata: geopandas.GeoDataFrame
        See :class:`pywatemsedem.geo.vectors.AbstractVector`
    geometry_type: str
        See :class:`pywatemsedem.geo.vectors.AbstractVector`
    req_geometry_type: str, default None
        See :class:`pywatemsedem.geo.vectors.AbstractVector`
    allow_empty: bool, default False
        See :class:`pywatemsedem.geo.vectors.AbstractVector`
    """

    def __init__(
        self, geodata, geometry_type, req_geometry_type=None, allow_empty=False
    ):
        """Initialize RasterMemory"""
        super().__init__(
            geodata, geometry_type, req_geometry_type, allow_empty=allow_empty
        )

    def clip(self):
        """NotImplemented clip function"""
        raise NotImplementedError


class VectorFile(AbstractVector):
    """clipped Vector based on input vector file"""

    def __init__(
        self, file_path, req_geometry_type=None, vct_clip=None, allow_empty=False
    ):
        """Initialize RasterFile class

        Parameters
        ----------
        file_path: pathlib.Path
            File path to user input raster.
        vct_clip: pathlib.Path
            Mask vector
        req_geometry_type: str
            Required type of geometry, see implemented geometries in
            :func:`pywatemsedem.geo.vectors.AbstractVector.check_type`
        allow_empty: bool, default False
            See :class:`pywatemsedem.geo.vectors.AbstractVector`
        """
        geometry_type = get_geometry_type(file_path)

        if vct_clip is not None:
            geodata = self.clip(file_path, vct_clip)
        else:
            geodata = gpd.read_file(file_path)

        super().__init__(
            geodata,
            geometry_type,
            req_geometry_type=req_geometry_type,
            allow_empty=allow_empty,
        )

    def clip(self, file_path, vct_clip):
        """Clip input file path with vct_clip

        Parameters
        ----------
        file_path: pathlib.Path
            File path to user input raster.
        vct_clip: pathlib.Path
            Mask vector
        geometry_type: str
            Required type of geometry, see implemented geometries in
            :func:`pywatemsedem.geo.vectors.AbstractVector.check_type`

        Returns
        -------
        geopandas.GeoDataFrame
        """
        vct_temp = Path(
            tempfile.NamedTemporaryFile(suffix=".shp", prefix=PREFIX_TEMP).name
        )
        clip_vct(file_path, vct_temp, vct_clip)
        geodata = gpd.read_file(vct_temp)
        clean_up_tempfiles(vct_temp, "shp")

        return geodata



import numpy as np

from pywatemsedem.io.valid import valid_boundaries


class PywatemsedemVectorAttributeError(Exception):
    """Raise error when input data are not conform the required pywatemsedem attributes."""


class PywatemsedemVectorAttributeValueError(Exception):
    """Raise error when input data attribute values are not conform required attributes
    values"""


class PywatemsedemRasterValueError(Exception):
    """Raise error when input data attribute values are not conform required
    attributes
    values"""


def raster_discrete_value_error(raster_array, tag, allowed_values, classes=None):
    """Value error for discrete attributes in input raster

    Raises error when an attribute of an input raster contains incorrect discrete
    values.

    Parameters
    ----------
    raster_array: numpy.ndarray
        Input raster.
    tag: str
        Internal naming.
    allowed_values: list
        Allowed values
    classes: list
        Name of classes of values, if None: not considered.
    """
    if not set(np.unique(raster_array)).issubset(allowed_values):
        if classes is not None:
            allowed_values = [f"{x}: {y}" for x, y in zip(allowed_values, classes)]
        str_values = "'" + "' ,'".join(allowed_values) + "'"
        msg = f"Input raster for '{tag}' can only contain values {str_values}"
        raise PywatemsedemRasterValueError(msg)


def attribute_continuous_value_error(source, tag, attribute, lower=None, upper=None):
    """Value error for continuous attributes in input vector

    Raises error when an attribute of an input vector contains out-of-bound values.

    Parameters
    ----------
    source: geopandas.GeoDataFrame
        Input vector.
    tag: str
        Internal naming.
    lower: int or float, default None
        Lower boundary, if None: no boundary.
    upper: int or float, default None
        Upper boundary, if None: no boundary.
    """
    try:
        valid_boundaries(source[attribute], lower=lower, upper=upper)
    except ValueError:
        if lower is None:
            lower = -float("inf")
        if upper is None:
            upper = float("inf")
        msg = (
            f"Column '{attribute}' in '{tag}' vector should contain values in "
            f"[{lower}, {upper}]"
        )
        raise PywatemsedemVectorAttributeValueError(msg)


def attribute_discrete_value_error(
    source, tag, attribute, allowed_values, classes=None
):
    """Value error for discrete attributes in input vector

    Raises error when an attribute of an input vector contains incorrect discrete
    values.

    Parameters
    ----------
    source: geopandas.GeoDataFrame
        Input vector.
    tag: str
        Internal naming.
    attribute: str
        Column name
    allowed_values: set of int and/or float
        Allowed values
    classes: set of string, default None
        Name of classes of values, if None: not considered.
    """
    if not set(np.unique(source[attribute])).issubset(allowed_values):
        if classes is not None:
            allowed_values = [f"{x}: {y}" for x, y in zip(allowed_values, classes)]
        str_values = "'" + "' ,'".join(allowed_values) + "'"
        msg = (
            f"Column '{attribute}' in '{tag}' vector can only contain values "
            f"{str_values}"
        )
        raise PywatemsedemVectorAttributeValueError(msg)


def missing_attribute_error_in_vct(source, tag, req_attributes):
    """Missing attribute error in vectors

    Raises error when not all required attributes/columns are present in input vector.

    Parameters
    ----------
    source: geopandas.GeoDataFrame
        Input vector.
    tag: str
        Internal naming.
    req_attributes: set
        Required attributes in input source.
    """

    if not req_attributes.issubset(set(source.columns)):
        str_col = "'" + "' ,'".join(req_attributes) + "'"
        msg = (
            f"'{tag}' input vector should contain {str_col} as vector attribute/column."
        )
        raise PywatemsedemVectorAttributeError(msg)
